# copyright ################################# #
# This file is part of the wakis Package.     #
# Copyright (c) CERN, 2024.                   #
# ########################################### #


import copy

import numpy as xp

try:
    import cupy as xp_gpu

    imported_cupy = True
except ImportError:
    imported_cupy = False


class Field:
    """
    Class to handle 3D vector fields stored in a flattened 1D array.
    Uses lexicographic numbering:
        n = 1 + (i-1) + (j-1)*Nx + (k-1)*Nx*Ny
        len(n) = Nx*Ny*Nz

    Parameters
    ----------
    Nx : int
        Number of grid points in the x direction.
    Ny : int
        Number of grid points in the y direction.
    Nz : int
        Number of grid points in the z direction.
    dtype : type, optional
        Data type of the field array. Default is float.
    use_ones : bool, optional
        If True, initialize the field array with ones. Otherwise, zeros. Default is False.
    use_gpu : bool, optional
        If True, use CuPy for GPU arrays. Default is False.

    Attributes
    ----------
    Nx, Ny, Nz : int
        Grid dimensions.
    N : int
        Total number of grid points (Nx * Ny * Nz).
    dtype : type
        Data type of the field array.
    on_gpu : bool
        Whether the field is stored on GPU.
    xp : module
        Numpy or cupy module, depending on use_gpu.
    array : ndarray
        Flattened 1D field array of shape (N*3,).
    """

    def __init__(self, Nx, Ny, Nz, dtype=float, use_ones=False, use_gpu=False):
        self.Nx = Nx
        self.Ny = Ny
        self.Nz = Nz
        self.N = Nx * Ny * Nz
        self.dtype = dtype
        self.on_gpu = use_gpu

        if self.on_gpu:
            if imported_cupy:
                self.xp = xp_gpu
            else:
                print(
                    "*** cupy could not be imported, please CUDA check installation"
                )
        else:
            self.xp = xp

        if use_ones:
            self.array = self.xp.ones(self.N * 3, dtype=self.dtype, order="F")
        else:
            self.array = self.xp.zeros(self.N * 3, dtype=self.dtype, order="F")

    @property
    def field_x(self):
        """Return the x-component of the field as a 1D array."""
        return self.array[0 : self.N]

    @property
    def field_y(self):
        """Return the y-component of the field as a 1D array."""
        return self.array[self.N : 2 * self.N]

    @property
    def field_z(self):
        """Return the z-component of the field as a 1D array."""
        return self.array[2 * self.N : 3 * self.N]

    @field_x.setter
    def field_x(self, value):
        """Set the x-component of the field."""
        if len(value.shape) > 1:
            self.from_matrix(value, "x")
        else:
            self.array[0 : self.N] = value

    @field_y.setter
    def field_y(self, value):
        """Set the y-component of the field."""
        if len(value.shape) > 1:
            self.from_matrix(value, "y")
        else:
            self.array[self.N : 2 * self.N] = value

    @field_z.setter
    def field_z(self, value):
        """Set the z-component of the field."""
        if len(value.shape) > 1:
            self.from_matrix(value, "z")
        else:
            self.array[2 * self.N : 3 * self.N] = value

    def toarray(self):
        """
        Return the flattened field array.

        Returns
        -------
        array : ndarray
            Flattened field array of shape (N*3,).
        """
        return self.array

    def fromarray(self, array):
        """
        Set the field array from a flattened array.

        Parameters
        ----------
        array : ndarray
            Flattened field array of shape (N*3,).
        """
        self.array[:] = array

    def to_matrix(self, key):
        """
        Return the specified component as a 3D matrix.

        Parameters
        ----------
        key : int or str
            Component to return: 0 or 'x', 1 or 'y', 2 or 'z'.

        Returns
        -------
        mat : ndarray
            3D array of shape (Nx, Ny, Nz) for the selected component.
        """
        if key == 0 or key == "x":
            return self.xp.reshape(
                self.array[0 : self.N], (self.Nx, self.Ny, self.Nz), order="F"
            )
        if key == 1 or key == "y":
            return self.xp.reshape(
                self.array[self.N : 2 * self.N],
                (self.Nx, self.Ny, self.Nz),
                order="F",
            )
        if key == 2 or key == "z":
            return self.xp.reshape(
                self.array[2 * self.N : 3 * self.N],
                (self.Nx, self.Ny, self.Nz),
                order="F",
            )

    def from_matrix(self, mat, key):
        """
        Set the specified component from a 3D matrix.

        Parameters
        ----------
        mat : ndarray
            3D array of shape (Nx, Ny, Nz).
        key : int or str
            Component to set: 0 or 'x', 1 or 'y', 2 or 'z'.
        """
        if key == 0 or key == "x":
            self.array[0 : self.N] = self.xp.reshape(mat, self.N, order="F")
        elif key == 1 or key == "y":
            self.array[self.N : 2 * self.N] = self.xp.reshape(
                mat, self.N, order="F"
            )
        elif key == 2 or key == "z":
            self.array[2 * self.N : 3 * self.N] = self.xp.reshape(
                mat, self.N, order="F"
            )
        else:
            raise IndexError("Component id not valid")

    def to_gpu(self):
        """
        Move the field array to GPU (CuPy).

        Returns
        -------
        None
        """
        if imported_cupy:
            self.xp = xp_gpu
            self.array = self.xp.asarray(self.array)  # to cupy arr
            self.on_gpu = True
        else:
            print("*** CuPy is not imported")
            pass

    def from_gpu(self):
        """
        Move the field array from GPU (CuPy) to CPU (NumPy).

        Returns
        -------
        None
        """
        if self.on_gpu:
            self.array = self.array.get()  # to numpy arr
            self.on_gpu = False
        else:
            print("*** GPU is not enabled")
            pass

    def __getitem__(self, key):
        """
        Get a field value or slice.

        Parameters
        ----------
        key : tuple, int, or slice
            - (ix, iy, iz, component): 3D index and component ('x', 'y', 'z', or 'Abs')
            - int: lexico-graphic index
            - slice: slice of the flattened array

        Returns
        -------
        value : float or ndarray
            Field value(s) at the specified location.
        """
        if type(key) is tuple:
            if len(key) != 4:
                raise IndexError(
                    "Need 3 indexes and component to access the field"
                )
            if key[3] == 0 or key[3] == "x":
                if self.on_gpu:
                    field = self.xp.reshape(
                        self.array[0 : self.N],
                        (self.Nx, self.Ny, self.Nz),
                        order="F",
                    )
                    return field[key[0], key[1], key[2]].get()
                else:
                    field = self.xp.reshape(
                        self.array[0 : self.N],
                        (self.Nx, self.Ny, self.Nz),
                        order="F",
                    )
                    return field[key[0], key[1], key[2]]
            elif key[3] == 1 or key[3] == "y":
                if self.on_gpu:
                    field = self.xp.reshape(
                        self.array[self.N : 2 * self.N],
                        (self.Nx, self.Ny, self.Nz),
                        order="F",
                    )
                    return field[key[0], key[1], key[2]].get()
                else:
                    field = self.xp.reshape(
                        self.array[self.N : 2 * self.N],
                        (self.Nx, self.Ny, self.Nz),
                        order="F",
                    )
                    return field[key[0], key[1], key[2]]
            elif key[3] == 2 or key[3] == "z":
                if self.on_gpu:
                    field = self.xp.reshape(
                        self.array[2 * self.N : 3 * self.N],
                        (self.Nx, self.Ny, self.Nz),
                        order="F",
                    )
                    return field[key[0], key[1], key[2]].get()
                else:
                    field = self.xp.reshape(
                        self.array[2 * self.N : 3 * self.N],
                        (self.Nx, self.Ny, self.Nz),
                        order="F",
                    )
                    return field[key[0], key[1], key[2]]
            elif type(key[3]) is str and key[3].lower() == "abs":
                field = self.get_abs()
                return field[key[0], key[1], key[2]]
            else:
                raise IndexError("Component id not valid")

        elif type(key) is int:
            if key <= self.N:
                if self.on_gpu:
                    return self.array[key].get()
                else:
                    return self.array[key]
            else:
                raise IndexError(
                    "Lexico-graphic index cannot be higher than product of dimensions"
                )

        elif type(key) is slice:
            if self.on_gpu:
                return self.array[key].get()
            else:
                return self.array[key]

        else:
            raise ValueError("key must be a 3-tuple or an integer")

    def __setitem__(self, key, value):
        """
        Set a field value or slice.

        Parameters
        ----------
        key : tuple, int, or slice
            - (ix, iy, iz, component): 3D index and component
            - int: lexico-graphic index
            - slice: slice of the flattened array
        value : float or ndarray
            Value(s) to set.
        """
        if self.on_gpu:
            value = self.xp.asarray(value)

        if type(key) is tuple:
            if len(key) != 4:
                raise IndexError(
                    "Need 3 indexes and component to access the field"
                )
            else:
                field = self.to_matrix(key[3])
                field[key[0], key[1], key[2]] = value
                self.from_matrix(field, key[3])

        elif type(key) is int:
            if key <= self.N:
                self.array[key] = value
            else:
                raise IndexError(
                    "Lexico-graphic index cannot be higher than product of dimensions"
                )

        elif type(key) is slice:
            self.array[key] = value
        else:
            raise IndexError("key must be a 3-tuple or an integer")

    def __mul__(self, other, dtype=None):
        """
        Multiply the field by a scalar, array, or matrix.

        Parameters
        ----------
        other : float, int, ndarray
            Scalar, 1D array, or 3D matrix to multiply.
        dtype : type, optional
            Data type for the result.

        Returns
        -------
        Field
            New Field object with the result.
        """
        if dtype is None:
            dtype = self.dtype

        # other is number
        if type(other) is float or type(other) is int:
            mulField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            mulField.array = self.array * other

        # other is matrix
        elif len(other.shape) > 1:
            mulField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            for d in ["x", "y", "z"]:
                mulField.from_matrix(self.to_matrix(d) * other, d)

        # other is 1d array
        else:
            mulField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            mulField.array = self.array * other

        return mulField

    def __div__(self, other, dtype=None):
        """
        Divide the field by a scalar, array, or matrix.

        Parameters
        ----------
        other : float, int, ndarray
            Scalar, 1D array, or 3D matrix to divide by.
        dtype : type, optional
            Data type for the result.

        Returns
        -------
        Field
            New Field object with the result.
        """
        if dtype is None:
            dtype = self.dtype

        # other is number
        if type(other) is float or type(other) is int:
            divField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            divField.array = self.array / other

        # other is matrix
        if len(other.shape) > 1:
            divField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            for d in ["x", "y", "z"]:
                divField.from_matrix(self.to_matrix(d) / other, d)

        # other is constant or 1d array
        else:
            divField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            divField.array = self.array / other

        return divField

    def __add__(self, other, dtype=None):
        """
        Add another field, scalar, array, or matrix to this field.

        Parameters
        ----------
        other : Field, float, int, ndarray
            Field, scalar, 1D array, or 3D matrix to add.
        dtype : type, optional
            Data type for the result.

        Returns
        -------
        Field
            New Field object with the result.
        """
        if dtype is None:
            dtype = self.dtype

        if type(other) is Field:
            addField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            addField.field_x = self.field_x + other.field_x
            addField.field_y = self.field_y + other.field_y
            addField.field_z = self.field_z + other.field_z

        # other is matrix
        elif len(other.shape) > 1:
            addField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            for d in ["x", "y", "z"]:
                addField.from_matrix(self.to_matrix(d) + other, d)

        # other is constant or 1d array
        else:
            addField = Field(self.Nx, self.Ny, self.Nz, dtype=dtype)
            addField.array = self.array + other

        return addField

    def __repr__(self):
        """String representation of the field (for debugging)."""
        return (
            "x:\n"
            + self.field_x.__repr__()
            + "\n"
            + "y:\n"
            + self.field_y.__repr__()
            + "\n"
            + "z:\n"
            + self.field_z.__repr__()
        )

    def __str__(self):
        """String representation of the field."""
        return (
            "x:\n"
            + self.field_x.__str__()
            + "\n"
            + "y:\n"
            + self.field_y.__str__()
            + "\n"
            + "z:\n"
            + self.field_z.__str__()
        )

    def copy(self):
        """
        Return a deep copy of the Field object.

        Returns
        -------
        Field
            Deep copy of the field.
        """
        obj = type(self).__new__(self.__class__)  # Create empty instance

        for key, value in self.__dict__.items():
            if key == "xp":
                obj.xp = self.xp  # Just copy reference, no need for deepcopy
            elif key == "array" and self.on_gpu:
                obj.array = self.xp.array(
                    self.array
                )  # Ensure CuPy array is copied properly
            else:
                obj.__dict__[key] = copy.deepcopy(value)

        return obj

    def compute_ijk(self, n):
        """
        Compute (i, j, k) indices from a lexico-graphic index.

        Parameters
        ----------
        n : int
            Lexico-graphic index.

        Returns
        -------
        i, j, k : int
            3D indices corresponding to the lexico-graphic index.
        """
        if n > (self.N):
            raise IndexError(
                "Lexico-graphic index cannot be higher than product of dimensions"
            )

        k = n // (self.Nx * self.Ny)
        i = (n - k * self.Nx * self.Ny) % self.Nx
        j = (n - k * self.Nx * self.Ny) // self.Nx

        return i, j, k

    def get_abs(self, as_matrix=True):
        """
        Compute the magnitude of the field.

        Parameters
        ----------
        as_matrix : bool, optional
            If True, return as a 3D matrix. If False, return as a 1D array.

        Returns
        -------
        abs_field : ndarray
            Magnitude of the field.
        """
        if as_matrix:
            if self.on_gpu:
                return xp.sqrt(
                    self.to_matrix("x") ** 2
                    + self.to_matrix("y") ** 2
                    + self.to_matrix("z") ** 2
                ).get()
            else:
                return xp.sqrt(
                    self.to_matrix("x") ** 2
                    + self.to_matrix("y") ** 2
                    + self.to_matrix("z") ** 2
                )

        else:  # 1d array
            if self.on_gpu:
                return xp.sqrt(
                    self.field_x**2 + self.field_y**2 + self.field_z**2
                ).get()
            else:
                return xp.sqrt(
                    self.field_x**2 + self.field_y**2 + self.field_z**2
                )

    def inspect(
        self,
        plane="YZ",
        cmap="bwr",
        dpi=100,
        figsize=[8, 6],
        x=None,
        y=None,
        z=None,
        off_screen=False,
        handles=False,
        **kwargs,
    ):
        """
        Visualize 2D slices of the field components using matplotlib.

        Parameters
        ----------
        plane : {'XY', 'XZ', 'YZ'}, optional
            Plane to visualize. Default is 'YZ'.
        cmap : str, optional
            Colormap for the plot. Default is 'bwr'.
        dpi : int, optional
            Figure DPI. Default is 100.
        figsize : list, optional
            Figure size. Default is [8, 6].
        x, y, z : int, slice, or None, optional
            Custom slice indices. If all are not None, use as custom slice.
        off_screen : bool, optional
            If True, display the plot off-screen. Default is False.
        handles : bool, optional
            If True, return (fig, axs) instead of showing. Default is False.
        **kwargs
            Additional keyword arguments for imshow.

        Returns
        -------
        fig, axs : tuple, optional
            Returned if handles=True.
        None
            Otherwise.
        """
        import matplotlib.pyplot as plt
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        if None not in (x, y, z):  # custom slice
            transpose = False
            extent = None
            xax, yax = "No. of cells", "No. of cells"

        elif plane == "XY":
            key = [slice(0, self.Nx), slice(0, self.Ny), int(self.Nz // 2)]
            x, y, z = key[0], key[1], key[2]
            extent = (0, self.Nx, 0, self.Ny)
            xax, yax = "nx", "ny"
            transpose = True

        elif plane == "XZ":
            key = [slice(0, self.Nx), int(self.Ny // 2), slice(0, self.Nz)]
            x, y, z = key[0], key[1], key[2]
            extent = (0, self.Nz, 0, self.Nx)
            xax, yax = "nz", "nx"
            transpose = False

        elif plane == "YZ":
            key = [int(self.Nx // 2), slice(0, self.Ny), slice(0, self.Nz)]
            x, y, z = key[0], key[1], key[2]
            extent = (0, self.Nz, 0, self.Ny)
            xax, yax = "nz", "ny"
            transpose = False

        fig, axs = plt.subplots(
            1, 3, tight_layout=True, figsize=figsize, dpi=dpi
        )
        dims = {0: "x", 1: "y", 2: "z"}

        im = {}

        for d in [0, 1, 2]:
            field = self.to_matrix(d)

            if self.on_gpu and hasattr(field, "get"):
                field = field.get()

            if transpose:
                im[d] = axs[d].imshow(
                    field[x, y, z].T,
                    cmap=cmap,
                    vmin=-field.max(),
                    vmax=field.max(),
                    extent=extent,
                    origin="lower",
                    **kwargs,
                )

            else:
                im[d] = axs[d].imshow(
                    field[x, y, z],
                    cmap=cmap,
                    vmin=-field.max(),
                    vmax=field.max(),
                    extent=extent,
                    origin="lower",
                    **kwargs,
                )

        for i, ax in enumerate(axs):
            ax.set_title(f"Field {dims[i]}, plane {plane}")
            fig.colorbar(
                im[i],
                cax=make_axes_locatable(ax).append_axes(
                    "right", size="5%", pad=0.1
                ),
            )
            ax.set_xlabel(xax)
            ax.set_ylabel(yax)

        if handles:
            return fig, axs

        if not off_screen:
            plt.show()
            return None

    def inspect3D(
        self,
        field="all",
        backend="pyista",
        grid=None,
        xmax=None,
        ymax=None,
        zmax=None,
        bounding_box=True,
        show_grid=True,
        cmap="viridis",
        dpi=100,
        off_screen=False,
        handles=False,
    ):
        """
        Visualize 3D field data on the structured grid using either Matplotlib
        (voxel rendering) or PyVista (interactive clipping and slicing).

        This method provides two complementary visualization backends:
        - **Matplotlib**: static voxel plots of the field components (x, y, z)
          or all combined, useful for quick inspection, but memory intensive.
        - **PyVista**: interactive 3D visualization with sliders to dynamically
          clip the volume along X, Y, and Z, and optional wireframe slices.

        Parameters
        ----------
        field : {'x', 'y', 'z', 'all'}, optional
            Which field component(s) to visualize. Default is 'all'.
        backend : {'matplotlib', 'pyvista'}, optional
            Visualization backend to use. Default is 'pyvista'.
        grid : object, optional
            Structured grid object to use for visualization. If None, a grid is
            constructed from the solver's internal dimensions.
        xmax, ymax, zmax : int or float, optional
            Maximum extents in each direction for visualization. Defaults to the
            full grid dimensions if not specified.
        bounding_box : bool, optional
            If True, draw a wireframe bounding box of the simulation domain
            (only used in PyVista backend). Default is True.
        show_grid : bool, optional
            If True, show wireframe slice planes of the grid during interactive
            visualization (PyVista backend). Default is True.
        cmap : str, optional
            Colormap to apply to the scalar field. Default is 'viridis'.
        dpi : int, optional
            Resolution of Matplotlib figures (only for Matplotlib backend).
            Default is 100.
        show : bool, optional
            Whether to display the figure/plot immediately. If False in PyVista,
            exports to `field.html` instead. Default is True.
        handles : bool, optional
            If True, return figure/axes (Matplotlib) or the Plotter object
            (PyVista) for further customization instead of showing directly.
            Default is False.

        Returns
        -------
        fig, axs : tuple, optional
            Returned when `backend='matplotlib'` and `handles=True`.
        pl : pyvista.Plotter, optional
            Returned when `backend='pyvista'` and `handles=True`.

        Notes
        -----
        - The PyVista backend provides interactive sliders to clip the
          volume along each axis independently and inspect internal
          structures of the 3D field.
        - The Matplotlib backend provides a quick static voxel rendering
          but is limited in interactivity and scalability.
        """

        field = field.lower()

        # ---------- matplotlib backend ---------------
        if backend.lower() == "matplotlib":
            import matplotlib as mpl
            import matplotlib.pyplot as plt

            fig = plt.figure(tight_layout=True, dpi=dpi, figsize=[12, 6])

            plot_x, plot_y, plot_z = False, False, False

            if field == "all":
                plot_x = True
                plot_y = True
                plot_z = True

            elif field.lower() == "x":
                plot_x = True
            elif field.lower() == "y":
                plot_y = True
            elif field.lower() == "z":
                plot_z = True

            if xmax is None:
                xmax = self.Nx
            if ymax is None:
                ymax = self.Ny
            if zmax is None:
                zmax = self.Nz

            x, y, z = self.xp.mgrid[0 : xmax + 1, 0 : ymax + 1, 0 : zmax + 1]
            axs = []

            # field x
            if plot_x:
                arr = self.to_matrix("x")[
                    0 : int(xmax), 0 : int(ymax), 0 : int(zmax)
                ]
                if field == "all":
                    ax = fig.add_subplot(1, 3, 1, projection="3d")
                else:
                    ax = fig.add_subplot(1, 1, 1, projection="3d")

                vmin, vmax = (
                    -self.xp.max(self.xp.abs(arr)),
                    +self.xp.max(self.xp.abs(arr)),
                )
                norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
                colors = mpl.colormaps[cmap](norm(arr))
                ax.voxels(
                    x, y, z, filled=self.xp.ones_like(arr), facecolors=colors
                )

                m = mpl.cm.ScalarMappable(cmap=cmap, norm=norm)
                m.set_array([])
                fig.colorbar(m, ax=ax, shrink=0.5, aspect=10)
                ax.set_title("Field x")
                axs.append(ax)

            # field y
            if plot_y:
                arr = self.to_matrix("y")[
                    0 : int(xmax), 0 : int(ymax), 0 : int(zmax)
                ]
                if field == "all":
                    ax = fig.add_subplot(1, 3, 2, projection="3d")
                else:
                    ax = fig.add_subplot(1, 1, 1, projection="3d")

                vmin, vmax = (
                    -self.xp.max(self.xp.abs(arr)),
                    +self.xp.max(self.xp.abs(arr)),
                )
                norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
                colors = mpl.colormaps[cmap](norm(arr))
                ax.voxels(
                    x, y, z, filled=self.xp.ones_like(arr), facecolors=colors
                )

                m = mpl.cm.ScalarMappable(cmap=cmap, norm=norm)
                m.set_array([])
                fig.colorbar(m, ax=ax, shrink=0.5, aspect=10)
                ax.set_title("Field y")
                axs.append(ax)

            # field z
            if plot_z:
                arr = self.to_matrix("z")[
                    0 : int(xmax), 0 : int(ymax), 0 : int(zmax)
                ]
                if field == "all":
                    ax = fig.add_subplot(1, 3, 3, projection="3d")
                else:
                    ax = fig.add_subplot(1, 1, 1, projection="3d")

                vmin, vmax = (
                    -self.xp.max(self.xp.abs(arr)),
                    +self.xp.max(self.xp.abs(arr)),
                )
                norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
                colors = mpl.colormaps[cmap](norm(arr))
                ax.voxels(
                    x, y, z, filled=self.xp.ones_like(arr), facecolors=colors
                )

                m = mpl.cm.ScalarMappable(cmap=cmap, norm=norm)
                m.set_array([])
                fig.colorbar(m, ax=ax, shrink=0.5, aspect=10)
                ax.set_title("Field z")
                axs.append(ax)

            for i, ax in enumerate(axs):
                ax.set_xlabel("Nx")
                ax.set_ylabel("Ny")
                ax.set_zlabel("Nz")
                ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
                ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
                ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
                ax.set_xlim(self.Nx, 0)
                ax.set_ylim(self.Ny, 0)
                ax.set_zlim(self.Nz, 0)

            if handles:
                return fig, axs

            if not off_screen:
                plt.show()

        # ----------- pyvista backend ---------------
        else:
            import pyvista as pv

            if grid is not None and hasattr(grid, "grid"):
                xlo, xhi, ylo, yhi, zlo, zhi = (
                    grid.xmin,
                    grid.xmax,
                    grid.ymin,
                    grid.ymax,
                    grid.zmin,
                    grid.zmax,
                )
                grid = grid.grid
                if field == "x":
                    scalars = "Field " + field
                    grid[scalars] = xp.reshape(self.to_matrix(field), self.N)
                elif field == "y":
                    scalars = "Field " + field
                    grid[scalars] = xp.reshape(self.to_matrix(field), self.N)
                elif field == "z":
                    scalars = "Field " + field
                    grid[scalars] = xp.reshape(self.to_matrix(field), self.N)
                else:  # for all or abs
                    scalars = "Field " + "Abs"
                    grid[scalars] = xp.reshape(self.get_abs(), self.N)

                if xmax is None:
                    xmax = xhi
                if ymax is None:
                    ymax = yhi
                if zmax is None:
                    zmax = zhi

            else:
                print(
                    "[!] `grid` is not passed or is not a GridFIT3D object -> Using #N cells instead "
                )
                x = xp.linspace(0, self.Nx, self.Nx + 1)
                y = xp.linspace(0, self.Ny, self.Ny + 1)
                z = xp.linspace(0, self.Nz, self.Nz + 1)
                xlo, xhi, ylo, yhi, zlo, zhi = (
                    x.min(),
                    x.max(),
                    y.min(),
                    y.max(),
                    z.min(),
                    z.max(),
                )
                if xmax is None:
                    xmax = self.Nx
                if ymax is None:
                    ymax = self.Ny
                if zmax is None:
                    zmax = self.Nz
                X, Y, Z = xp.meshgrid(x, y, z, indexing="ij")
                grid = pv.StructuredGrid(
                    X.transpose(), Y.transpose(), Z.transpose()
                )

                if field == "x":
                    scalars = "Field " + field
                    grid[scalars] = xp.reshape(self.to_matrix(field), self.N)
                elif field == "y":
                    scalars = "Field " + field
                    grid[scalars] = xp.reshape(self.to_matrix(field), self.N)
                elif field == "z":
                    scalars = "Field " + field
                    grid[scalars] = xp.reshape(self.to_matrix(field), self.N)
                else:  # for all or abs
                    scalars = "Field " + "Abs"
                    grid[scalars] = xp.reshape(self.get_abs(), self.N)

            pv.global_theme.allow_empty_mesh = True
            pl = pv.Plotter()
            vals = {"x": xmax, "y": ymax, "z": zmax}

            # --- Update function ---
            def update_clip(val, axis="x"):
                vals[axis] = val
                # define bounds dynamically
                if axis == "x":
                    slice_obj = grid.slice(normal="x", origin=(val, 0, 0))
                elif axis == "y":
                    slice_obj = grid.slice(normal="y", origin=(0, val, 0))
                else:  # z
                    slice_obj = grid.slice(normal="z", origin=(0, 0, val))

                # add clipped volume (scalars)
                pl.add_mesh(
                    grid.clip_box(
                        bounds=(
                            xlo,
                            vals["x"],
                            ylo,
                            vals["y"],
                            zlo,
                            vals["z"],
                        ),
                        invert=False,
                    ),
                    scalars=scalars,
                    cmap=cmap,
                    name="clip",
                )

                # add slice wireframe (grid structure)
                if show_grid:
                    pl.add_mesh(
                        slice_obj,
                        style="wireframe",
                        color="grey",
                        name="slice",
                    )

            # --- Sliders (placed side-by-side vertically) ---
            pl.add_slider_widget(
                lambda value: update_clip(value, "x"),
                [xlo, xhi],
                value=xmax,
                title="X Clip",
                pointa=(0.8, 0.8),
                pointb=(0.95, 0.8),  # top-right
                style="modern",
            )

            pl.add_slider_widget(
                lambda value: update_clip(value, "y"),
                [ylo, yhi],
                value=ymax,
                title="Y Clip",
                pointa=(0.8, 0.6),
                pointb=(0.95, 0.6),  # middle-right
                style="modern",
            )

            pl.add_slider_widget(
                lambda value: update_clip(value, "z"),
                [zlo, zhi],
                value=zmax,
                title="Z Clip",
                pointa=(0.8, 0.4),
                pointb=(0.95, 0.4),  # lower-right
                style="modern",
            )

            # Camera orientation
            pl.camera_position = "zx"
            pl.camera.azimuth += 30
            pl.camera.elevation += 30
            pl.set_background("mistyrose", top="white")
            try:
                pl.add_logo_widget("../docs/img/wakis-logo-pink.png")
            except Exception:
                pass
            pl.add_axes()
            pl.enable_3_lights()
            pl.enable_anti_aliasing()

            if bounding_box:
                pl.add_mesh(
                    pv.Box(bounds=(xlo, xhi, ylo, yhi, zlo, zhi)),
                    style="wireframe",
                    color="black",
                    line_width=2,
                    name="domain_box",
                )

            if handles:
                return pl

            if off_screen:
                pl.export_html("Field.html")
            else:
                pl.show()



import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
from scipy.constants import c

sys.path.append("../wakis")


from wakis import GridFIT3D, SolverFIT3D
from wakis.sources import PlaneWave


class TestPlanewave:
    img_folder = "tests/003_img/"

    def test_simulation(self, flag_offscreen):
        print("\n---------- Initializing simulation ------------------")
        # Number of mesh cells
        Nx = 60
        Ny = 60
        Nz = 120

        # Embedded boundaries
        stl_file = "tests/stl/003_sphere.stl"
        surf = pv.read(stl_file)

        stl_solids = {"Sphere": stl_file}
        stl_materials = {"Sphere": [10.0, 1.0]}  # dielectric [eps_r, mu_r]
        stl_rotate = [0, 0, 0]
        stl_scale = 1e-3

        surf = surf.rotate_x(stl_rotate[0])
        surf = surf.rotate_y(stl_rotate[1])
        surf = surf.rotate_z(stl_rotate[2])
        surf = surf.scale(stl_scale)

        # Domain bounds and grid
        xmin, xmax, ymin, ymax, zmin, zmax = surf.bounds
        padx, pady, padz = (
            (xmax - xmin) * 0.2,
            (ymax - ymin) * 0.2,
            (zmax - zmin) * 1.0,
        )

        xmin, ymin, zmin = (xmin - padx), (ymin - pady), (zmin - padz)
        xmax, ymax, zmax = (xmax + padx), (ymax + pady), (zmax + padz)

        global grid
        grid = GridFIT3D(
            xmin,
            xmax,
            ymin,
            ymax,
            zmin,
            zmax,
            Nx,
            Ny,
            Nz,
            stl_solids=stl_solids,
            stl_rotate=stl_rotate,
            stl_scale=stl_scale,
            stl_materials=stl_materials,
        )

        # Boundary conditions and
        bc_low = ["periodic", "periodic", "pec"]
        bc_high = ["periodic", "periodic", "pml"]

        # -------------- Output folder ---------------------
        if not os.path.exists(self.img_folder):
            os.mkdir(self.img_folder)

        # simulation
        global solver
        solver = SolverFIT3D(
            grid,
            use_stl=True,
            bc_low=bc_low,
            bc_high=bc_high,
            dtype=np.float32,
        )

        # source
        f = 15 / ((solver.z.max() - solver.z.min()) / c)
        source = PlaneWave(
            xs=slice(1, Nx - 1), ys=slice(1, Ny - 1), zs=1, f=f, beta=1.0
        )

        Nt = int(1.0 * (solver.z.max() - solver.z.min()) / c / solver.dt)
        solver.emsolve(Nt, source)

        if not flag_offscreen:
            plt.ion()

    def test_field_inspect(self, flag_offscreen):
        global solver
        # Inspect plane
        solver.E.inspect(
            plane="YZ",
            cmap="bwr",
            dpi=100,
            figsize=[8, 6],
            off_screen=flag_offscreen,
        )
        # Inspect custom slice
        solver.ieps.inspect(
            x=slice(20, 60),
            y=slice(20, 60),
            z=int(grid.Nz / 2),
            off_screen=flag_offscreen,
        )
        # Inspect with handles
        _, _ = solver.H.inspect(
            plane="XY", cmap="bwr", handles=True, off_screen=flag_offscreen
        )

    def test_plot1D(self, flag_offscreen):
        global solver
        solver.plot1D(
            "Ex",
            line="z",
            pos=[0.7, 0.6, 0.5, 0.4, 0.3, 0.2],
            xscale="linear",
            yscale="linear",
            off_screen=flag_offscreen,
            n=solver.Nt,
            colors=[
                "#5ccfe6",
                "#fdb6d0",
                "#ffae57",
                "#bae67e",
                "#ffd580",
                "#a2aabc",
            ],
            title=self.img_folder + "1Dplot_Ex",
        )

    def test_plot2D(self, flag_offscreen):
        global solver
        solver.plot2D(
            "Hy",
            plane="ZY",
            pos=0.5,
            cmap="bwr",
            add_patch="Sphere",
            patch_alpha=0.1,
            interpolation="spline36",
            off_screen=flag_offscreen,
            n=solver.Nt,
            title=self.img_folder + "2Dplot_Hy",
        )

    def test_plot3D_plane(self, flag_offscreen):
        global solver
        solver.plot3D(
            field="E",
            component="x",
            cmap="jet",
            add_stl="Sphere",
            stl_opacity=0.1,
            stl_colors="white",
            clip_interactive=True,
            clip_normal="-y",
            off_screen=flag_offscreen,
            title=self.img_folder + "3Dplot_Ex",
        )

    def test_plot3D_box(self, flag_offscreen):
        global solver
        solver.plot3D(
            field="H",
            component="y",
            cmap="bwr",
            add_stl="Sphere",
            stl_opacity=0.1,
            stl_colors="white",
            clip_box=True,
            clip_bounds=None,
            off_screen=flag_offscreen,
            title=self.img_folder + "3Dplot_Hy",
        )

    def test_plot3DonSTL_interactive(self, flag_offscreen):
        global solver
        solver.plot3DonSTL(
            "Ex",
            cmap="jet",
            stl_with_field="Sphere",
            field_opacity=1.0,
            stl_transparent="Sphere",
            stl_opacity=0.3,
            stl_colors="white",
            clip_interactive=True,
            clip_normal="-y",
            off_screen=flag_offscreen,
            zoom=1.0,
        )

    def test_plot3DonSTL_offscreen(self, flag_offscreen):
        global solver
        solver.plot3DonSTL(
            "Ex",
            cmap="jet",
            stl_with_field="Sphere",
            field_opacity=1.0,
            stl_transparent=None,
            stl_opacity=0.0,
            stl_colors=None,
            off_screen=flag_offscreen,
            zoom=1.0,
            title=self.img_folder + "3DplotOnSTL_Hy",
        )



import numpy as np
from numba import jit


def seg_length(x_1, y_1, x_2, y_2):
    return np.linalg.norm(np.array([x_1 - x_2, y_1 - y_2]))


@jit("b1(f8, f8)", nopython=True)
def eq(a, b):
    tol = 1e-8
    return abs(a - b) < tol


@jit("b1(f8, f8)", nopython=True)
def neq(a, b):
    return not eq(a, b)


class Grid2D:
    """
    Class holding the grid info and the routines for cell extensions.
      Constructor arguments:
          - xmin,xmax,ymin,ymax: extent of the domain.
          - nx, ny: number of cells per direction
          - conductors: conductor object
          - sol_type: type of solver. 'FDTD' for staircased FDTD, 'DM' for Conformal Dey-Mittra FDTD,
                      'ECT' for Extended Cell Technique conformal FDTD
    """

    def __init__(self, xmin, xmax, ymin, ymax, nx, ny, conductors, sol_type):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.nx = nx
        self.ny = ny
        self.dx = (xmax - xmin) / nx
        self.dy = (ymax - ymin) / ny
        self.conductors = conductors

        self.l_x = np.zeros((nx, ny + 1))
        self.l_y = np.zeros((nx + 1, ny))
        self.S = np.zeros((nx, ny))
        self.S_stab = np.zeros((nx, ny))
        self.S_enl = np.zeros((nx, ny))
        self.S_red = np.zeros((nx, ny))
        self.flag_unst_cell = np.zeros((nx, ny), dtype=bool)
        self.flag_int_cell = np.zeros((nx, ny), dtype=bool)
        self.flag_bound_cell = np.zeros((nx, ny), dtype=bool)
        self.flag_avail_cell = np.zeros((nx, ny), dtype=bool)
        self.flag_ext_cell = np.zeros((nx, ny), dtype=bool)
        self.flag_intr_cell = np.zeros((nx, ny), dtype=bool)
        self.broken = np.zeros((self.nx, self.ny), dtype=bool)

        if (sol_type != "FDTD") and (sol_type != "DM") and (sol_type != "ECT"):
            raise ValueError(
                "sol_type must be:\n"
                + "\t'FDTD' for standard staircased FDTD\n"
                + "\t'DM' for Dey-Mittra conformal FDTD\n"
                + "\t'ECT' for Enlarged Cell Technique conformal FDTD"
            )

        if sol_type == "DM" or sol_type == "FDTD":
            self.compute_edges(
                in_conductor=self.conductors.in_conductor,
                intersec_x=self.conductors.intersec_x,
                intersec_y=self.conductors.intersec_y,
                l_x=self.l_x,
                l_y=self.l_y,
                dx=self.dx,
                dy=self.dy,
                nx=self.nx,
                ny=self.ny,
                xmin=self.xmin,
                ymin=self.ymin,
            )
            compute_areas(
                l_x=self.l_x,
                l_y=self.l_y,
                S=self.S,
                S_red=self.S_red,
                nx=self.nx,
                ny=self.ny,
                dx=self.dx,
                dy=self.dy,
            )
            mark_cells(
                l_x=self.l_x,
                l_y=self.l_y,
                nx=self.nx,
                ny=self.ny,
                dx=self.dx,
                dy=self.dy,
                S=self.S,
                flag_int_cell=self.flag_int_cell,
                S_stab=self.S_stab,
                flag_unst_cell=self.flag_unst_cell,
                flag_bound_cell=self.flag_bound_cell,
                flag_avail_cell=self.flag_avail_cell,
            )
        elif sol_type == "ECT":
            self.compute_edges(
                in_conductor=self.conductors.in_conductor,
                intersec_x=self.conductors.intersec_x,
                intersec_y=self.conductors.intersec_y,
                l_x=self.l_x,
                l_y=self.l_y,
                dx=self.dx,
                dy=self.dy,
                nx=self.nx,
                ny=self.ny,
                xmin=self.xmin,
                ymin=self.ymin,
            )
            compute_areas(
                l_x=self.l_x,
                l_y=self.l_y,
                S=self.S,
                S_red=self.S_red,
                nx=self.nx,
                ny=self.ny,
                dx=self.dx,
                dy=self.dy,
            )
            mark_cells(
                l_x=self.l_x,
                l_y=self.l_y,
                nx=self.nx,
                ny=self.ny,
                dx=self.dx,
                dy=self.dy,
                S=self.S,
                flag_int_cell=self.flag_int_cell,
                S_stab=self.S_stab,
                flag_unst_cell=self.flag_unst_cell,
                flag_bound_cell=self.flag_bound_cell,
                flag_avail_cell=self.flag_avail_cell,
            )
            # info about intruded cells (i,j,[(i_borrowing,j_borrowing,area_borrowing, )])
            self.borrowing = np.empty((nx, ny), dtype=object)

            for i in range(nx):
                for j in range(ny):
                    self.borrowing[i, j] = []
            self.flag_ext_cell = self.flag_unst_cell.copy()
            self.compute_extensions(
                nx=nx,
                ny=ny,
                S=self.S,
                flag_int_cell=self.flag_int_cell,
                S_stab=self.S_stab,
                S_enl=self.S_enl,
                S_red=self.S_red,
                flag_unst_cell=self.flag_unst_cell,
                flag_avail_cell=self.flag_avail_cell,
                flag_ext_cell=self.flag_ext_cell,
                flag_intr_cell=self.flag_intr_cell,
                borrowing=self.borrowing,
            )

    """
  Function to compute the length of the edges of the conformal grid.
    Inputs:
        - tol: an edge shorter than tol will be considered as of zero length
    """

    @staticmethod
    def compute_edges(
        tol=1e-8,
        in_conductor=None,
        intersec_x=None,
        intersec_y=None,
        l_x=None,
        l_y=None,
        dx=None,
        dy=None,
        nx=None,
        ny=None,
        xmin=None,
        ymin=None,
    ):
        """
        Notation:

              (x_3, y_3)----- l_x[i, j + 1] ---------
                  |                                  |
                  |                                  |
                 l_y[i, j]        (i, j)            l_y[i + 1, j]
                  |                                  |
                  |                                  |
              (x_1, y_1)------- l_x[i, j] -------(x_2, y_2)

        """
        for ii in range(nx):
            for jj in range(ny + 1):
                x_1 = ii * dx + xmin
                y_1 = jj * dy + ymin
                x_2 = (ii + 1) * dx + xmin
                y_2 = jj * dy + ymin
                # if point 1 is in conductor
                if in_conductor(x_1, y_1):
                    # if point 2 is in conductor, length of the l_x[i, j] is zero
                    if in_conductor(x_2, y_2):
                        l_x[ii, jj] = 0
                    # if point 2 is not in conductor, length of l_x[i, j] is the fractional length
                    else:
                        l_x[ii, jj] = seg_length(intersec_x(x_2, y_2), y_1, x_2, y_2)
                # if point 1 is not in conductor
                else:
                    # if point 2 is in conductor, length of l_x[i, j] is the fractional length
                    if in_conductor(x_2, y_2):
                        l_x[ii, jj] = seg_length(x_1, y_1, intersec_x(x_1, y_1), y_2)
                    # if point 2 is not in conductor, length of l_x[i, j] is dx
                    else:
                        l_x[ii, jj] = dx

        for ii in range(nx + 1):
            for jj in range(ny):
                x_1 = ii * dx + xmin
                y_1 = jj * dy + ymin
                x_3 = ii * dx + xmin
                y_3 = (jj + 1) * dy + ymin
                # if point 1 is in conductor
                if in_conductor(x_1, y_1):
                    # if point 3 to the right is in conductor, length of the l_y[i, j] is zero
                    if in_conductor(x_3, y_3):
                        l_y[ii, jj] = 0
                    # if point 3 is not in conductor, length of l_y[i, j] is the fractional length
                    else:
                        l_y[ii, jj] = seg_length(x_1, intersec_y(x_3, y_3), x_3, y_3)
                # if point 1 is not in conductor
                else:
                    # if point 3 is in conductor, length of the l_y[i, j] is the fractional length
                    if in_conductor(x_3, y_3):
                        l_y[ii, jj] = seg_length(x_1, y_1, x_3, intersec_y(x_1, y_1))
                    # if point 3 is not in conductor, length of l_y[i, j] is dy
                    else:
                        l_y[ii, jj] = dy

        # set to zero the length of very small cells
        if tol > 0.0:
            l_x /= dx
            l_y /= dy

            low_values_flags = abs(l_x) < tol
            high_values_flags = abs(l_x - 1.0) < tol
            l_x[low_values_flags] = 0
            l_x[high_values_flags] = 1.0

            low_values_flags = abs(l_y) < tol
            high_values_flags = abs(l_y - 1.0) < tol
            l_y[low_values_flags] = 0
            l_y[high_values_flags] = 1.0

            l_x *= dx
            l_y *= dy

    """
  Function to compute the extension of the unstable cells
    """

    @staticmethod
    def compute_extensions(
        nx=None,
        ny=None,
        S=None,
        flag_int_cell=None,
        S_stab=None,
        S_enl=None,
        S_red=None,
        flag_unst_cell=None,
        flag_avail_cell=None,
        flag_ext_cell=None,
        flag_intr_cell=None,
        borrowing=None,
        l_verbose=True,
        kk=0,
    ):
        N = np.sum(flag_ext_cell)

        if l_verbose:
            print(kk)
            print("ext cells: %d" % N)
        # Do the simple one-cell extension
        Grid2D._compute_extensions_one_cell(
            nx=nx,
            ny=ny,
            S=S,
            S_stab=S_stab,
            S_enl=S_enl,
            S_red=S_red,
            flag_avail_cell=flag_avail_cell,
            flag_ext_cell=flag_ext_cell,
            flag_intr_cell=flag_intr_cell,
            borrowing=borrowing,
        )

        N_one_cell = N - np.sum(flag_ext_cell)
        if l_verbose:
            print("one cell exts: %d" % N_one_cell)
        # If any cell could not be extended do the four-cell extension
        """
        if np.sum(flag_ext_cell) > 0:
            N = np.sum(flag_ext_cell)
            Grid2D._compute_extensions_four_cells(nx=nx, ny=ny, S=S, flag_int_cell=flag_int_cell,
                                                  S_stab=S_stab, S_enl=S_enl, S_red=S_red,
                                                  flag_unst_cell=flag_unst_cell,
                                                  flag_avail_cell=flag_avail_cell,
                                                  flag_ext_cell=flag_ext_cell,
                                                  flag_intr_cell = flag_intr_cell,
                                                  borrowing=borrowing)
            N_four_cells = (N - np.sum(flag_ext_cell))
            if 1: #l_verbose:
                print('four cell exts: %d' % N_four_cells)
        """
        # If any cell could not be extended do the eight-cell extension
        if np.sum(flag_ext_cell) > 0:
            N = np.sum(flag_ext_cell)
            Grid2D._compute_extensions_eight_cells(
                nx=nx,
                ny=ny,
                S=S,
                flag_int_cell=flag_int_cell,
                S_stab=S_stab,
                S_enl=S_enl,
                S_red=S_red,
                flag_unst_cell=flag_unst_cell,
                flag_avail_cell=flag_avail_cell,
                flag_ext_cell=flag_ext_cell,
                flag_intr_cell=flag_intr_cell,
                borrowing=borrowing,
            )
            N_eight_cells = N - np.sum(flag_ext_cell)
            if 1:  # l_verbose:
                print("eight cell exts: %d" % N_eight_cells)
        # If any cell could not be extended the algorithm failed
        if np.sum(flag_ext_cell) > 0:
            N = np.sum(flag_ext_cell)
            raise RuntimeError(
                str(N) + " cells could not be extended.\n" + "Please refine the mesh"
            )

    """
  Function to compute the one-cell extension of the unstable cells 
    """

    @staticmethod
    def _compute_extensions_one_cell(
        nx=None,
        ny=None,
        S=None,
        S_stab=None,
        S_enl=None,
        S_red=None,
        flag_avail_cell=None,
        flag_ext_cell=None,
        flag_intr_cell=None,
        borrowing=None,
    ):
        for ii in range(0, nx):
            for jj in range(0, ny):
                if flag_ext_cell[ii, jj]:
                    S_ext = S_stab[ii, jj] - S[ii, jj]
                    if S[ii - 1, jj] > S_ext and flag_avail_cell[ii - 1, jj]:
                        denom = S[ii - 1, jj]
                        patch = S_ext * S[ii - 1, jj] / denom
                        if S_red[ii - 1, jj] - patch > 0:
                            S_red[ii - 1, jj] -= patch
                            borrowing[ii, jj].append([ii - 1, jj, patch, None])
                            flag_intr_cell[ii - 1, jj] = True
                            S_enl[ii, jj] = S[ii, jj] + patch
                            flag_ext_cell[ii, jj] = False
                    if (
                        S[ii, jj - 1] > S_ext
                        and flag_avail_cell[ii, jj - 1]
                        and flag_ext_cell[ii, jj]
                    ):
                        denom = S[ii, jj - 1]
                        patch = S_ext * S[ii, jj - 1] / denom
                        if S_red[ii, jj - 1] - patch > 0:
                            S_red[ii, jj - 1] -= patch
                            borrowing[ii, jj].append([ii, jj - 1, patch, None])
                            flag_intr_cell[ii, jj - 1] = True
                            S_enl[ii, jj] = S[ii, jj] + patch
                            flag_ext_cell[ii, jj] = False
                    if (
                        S[ii, jj + 1] > S_ext
                        and flag_avail_cell[ii, jj + 1]
                        and flag_ext_cell[ii, jj]
                    ):
                        denom = S[ii, jj + 1]
                        patch = S_ext * S[ii, jj + 1] / denom
                        if S_red[ii, jj + 1] - patch > 0:
                            S_red[ii, jj + 1] -= patch
                            borrowing[ii, jj].append([ii, jj + 1, patch, None])
                            flag_intr_cell[ii, jj + 1] = True
                            S_enl[ii, jj] = S[ii, jj] + patch
                            flag_ext_cell[ii, jj] = False
                    if (
                        S[ii + 1, jj] > S_ext
                        and flag_avail_cell[ii + 1, jj]
                        and flag_ext_cell[ii, jj]
                    ):
                        denom = S[ii + 1, jj]
                        patch = S_ext * S[ii + 1, jj] / denom
                        if S_red[ii + 1, jj] - patch > 0:
                            S_red[ii + 1, jj] -= patch
                            borrowing[ii, jj].append([ii + 1, jj, patch, None])
                            flag_intr_cell[ii + 1, jj] = True
                            S_enl[ii, jj] = S[ii, jj] + patch
                            flag_ext_cell[ii, jj] = False

    """
  Function to compute the four-cell extension of the unstable cells 
    """

    @staticmethod
    def _compute_extensions_four_cells(
        nx=None,
        ny=None,
        S=None,
        flag_int_cell=None,
        S_stab=None,
        S_enl=None,
        S_red=None,
        flag_unst_cell=None,
        flag_avail_cell=None,
        flag_ext_cell=None,
        flag_intr_cell=None,
        borrowing=None,
    ):
        for ii in range(0, nx):
            for jj in range(0, ny):
                local_avail = flag_avail_cell.copy()
                if (
                    flag_unst_cell[ii, jj]
                    and flag_int_cell[ii, jj]
                    and flag_ext_cell[ii, jj]
                ):
                    denom = (
                        (flag_avail_cell[ii - 1, jj]) * S[ii - 1, jj]
                        + (flag_avail_cell[ii + 1, jj]) * S[ii + 1, jj]
                        + (flag_avail_cell[ii, jj - 1]) * S[ii, jj - 1]
                        + (flag_avail_cell[ii, jj + 1]) * S[ii, jj + 1]
                    )
                    S_ext = S_stab[ii, jj] - S[ii, jj]
                    neg_cell = True
                    # idea: if any cell would reach negative area it is locally not available.
                    #       then denom has to be recomputed from scratch

                    while denom >= S_ext and neg_cell:
                        neg_cell = False
                        if local_avail[ii - 1, jj]:
                            patch = S_ext * S[ii - 1, jj] / denom
                            if S_red[ii - 1, jj] - patch <= 0:
                                neg_cell = True
                                local_avail[ii - 1, jj] = False
                        if local_avail[ii + 1, jj]:
                            patch = S_ext * S[ii + 1, jj] / denom
                            if S_red[ii + 1, jj] - patch <= 0:
                                neg_cell = True
                                local_avail[ii + 1, jj] = False
                        if local_avail[ii, jj - 1]:
                            patch = S_ext * S[ii, jj - 1] / denom
                            if S_red[ii, jj - 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii, jj - 1] = False
                        if local_avail[ii, jj + 1]:
                            patch = S_ext * S[ii, jj + 1] / denom
                            if S_red[ii, jj + 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii, jj + 1] = False
                        denom = (
                            (local_avail[ii - 1, jj]) * S[ii - 1, jj]
                            + (local_avail[ii + 1, jj]) * S[ii + 1, jj]
                            + (local_avail[ii, jj - 1]) * S[ii, jj - 1]
                            + (local_avail[ii, jj + 1]) * S[ii, jj + 1]
                        )

                    # If possible, do 4-cell extension
                    if denom >= S_ext:
                        S_enl[ii, jj] = S[ii, jj]
                        if local_avail[ii - 1, jj]:
                            patch = S_ext * S[ii - 1, jj] / denom
                            borrowing[ii, jj].append([ii - 1, jj, patch, None])
                            flag_intr_cell[ii - 1, jj] = True
                            S_enl[ii, jj] += patch
                            S_red[ii - 1, jj] -= patch
                        if local_avail[ii + 1, jj]:
                            patch = S_ext * S[ii + 1, jj] / denom
                            borrowing[ii, jj].append([ii + 1, jj, patch, None])
                            flag_intr_cell[ii + 1, jj] = True
                            S_enl[ii, jj] += patch
                            S_red[ii + 1, jj] -= patch
                        if local_avail[ii, jj - 1]:
                            patch = S_ext * S[ii, jj - 1] / denom
                            borrowing[ii, jj].append([ii, jj - 1, patch, None])
                            flag_intr_cell[ii, jj - 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii, jj - 1] -= patch
                        if local_avail[ii, jj + 1]:
                            patch = S_ext * S[ii, jj + 1] / denom
                            borrowing[ii, jj].append([ii, jj + 1, patch, None])
                            flag_intr_cell[ii, jj + 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii, jj + 1] -= patch

                        flag_ext_cell[ii, jj] = False

    """
  Function to compute the eight-cell extension of the unstable cells 
    """

    @staticmethod
    def _compute_extensions_eight_cells(
        nx=None,
        ny=None,
        S=None,
        flag_int_cell=None,
        S_stab=None,
        S_enl=None,
        S_red=None,
        flag_unst_cell=None,
        flag_avail_cell=None,
        flag_ext_cell=None,
        flag_intr_cell=None,
        borrowing=None,
    ):
        for ii in range(0, nx):
            for jj in range(0, ny):
                local_avail = flag_avail_cell.copy()
                if (
                    flag_unst_cell[ii, jj]
                    and flag_int_cell[ii, jj]
                    and flag_ext_cell[ii, jj]
                ):
                    S_enl[ii, jj] = S[ii, jj]
                    S_ext = S_stab[ii, jj] - S[ii, jj]

                    denom = (
                        (flag_avail_cell[ii - 1, jj]) * S[ii - 1, jj]
                        + (flag_avail_cell[ii + 1, jj]) * S[ii + 1, jj]
                        + (flag_avail_cell[ii, jj - 1]) * S[ii, jj - 1]
                        + (flag_avail_cell[ii, jj + 1]) * S[ii, jj + 1]
                        + (flag_avail_cell[ii - 1, jj - 1]) * S[ii - 1, jj - 1]
                        + (flag_avail_cell[ii + 1, jj - 1]) * S[ii + 1, jj - 1]
                        + (flag_avail_cell[ii - 1, jj + 1]) * S[ii - 1, jj + 1]
                        + (flag_avail_cell[ii + 1, jj + 1]) * S[ii + 1, jj + 1]
                    )

                    neg_cell = True
                    while denom >= S_ext and neg_cell:
                        neg_cell = False
                        if local_avail[ii - 1, jj]:
                            patch = S_ext * S[ii - 1, jj] / denom
                            if S_red[ii - 1, jj] - patch <= 0:
                                neg_cell = True
                                local_avail[ii - 1, jj] = False
                        if local_avail[ii + 1, jj]:
                            patch = S_ext * S[ii + 1, jj] / denom
                            if S_red[ii + 1, jj] - patch <= 0:
                                neg_cell = True
                                local_avail[ii + 1, jj] = False
                        if local_avail[ii, jj - 1]:
                            patch = S_ext * S[ii, jj - 1] / denom
                            if S_red[ii, jj - 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii, jj - 1] = False
                        if local_avail[ii, jj + 1]:
                            patch = S_ext * S[ii, jj + 1] / denom
                            if S_red[ii, jj + 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii, jj + 1] = False
                        if local_avail[ii - 1, jj - 1]:
                            patch = S_ext * S[ii - 1, jj - 1] / denom
                            if S_red[ii - 1, jj - 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii - 1, jj - 1] = False
                        if local_avail[ii + 1, jj - 1]:
                            patch = S_ext * S[ii + 1, jj - 1] / denom
                            if S_red[ii + 1, jj - 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii + 1, jj - 1] = False
                        if local_avail[ii - 1, jj + 1]:
                            patch = S_ext * S[ii - 1, jj + 1] / denom
                            if S_red[ii - 1, jj + 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii - 1, jj + 1] = False
                        if local_avail[ii + 1, jj + 1]:
                            patch = S_ext * S[ii + 1, jj + 1] / denom
                            if S_red[ii + 1, jj + 1] - patch <= 0:
                                neg_cell = True
                                local_avail[ii + 1, jj + 1] = False

                        denom = (
                            (local_avail[ii - 1, jj]) * S[ii - 1, jj]
                            + (local_avail[ii + 1, jj]) * S[ii + 1, jj]
                            + (local_avail[ii, jj - 1]) * S[ii, jj - 1]
                            + (local_avail[ii, jj + 1]) * S[ii, jj + 1]
                            + (local_avail[ii - 1, jj - 1]) * S[ii - 1, jj - 1]
                            + (local_avail[ii + 1, jj - 1]) * S[ii + 1, jj - 1]
                            + (local_avail[ii - 1, jj + 1]) * S[ii - 1, jj + 1]
                            + (local_avail[ii + 1, jj + 1]) * S[ii + 1, jj + 1]
                        )

                    if denom >= S_ext:
                        S_enl[ii, jj] = S[ii, jj]
                        if local_avail[ii - 1, jj]:
                            patch = S_ext * S[ii - 1, jj] / denom
                            borrowing[ii, jj].append([ii - 1, jj, patch, None])
                            flag_intr_cell[ii - 1, jj] = True
                            S_enl[ii, jj] += patch
                            S_red[ii - 1, jj] -= patch
                        if local_avail[ii + 1, jj]:
                            patch = S_ext * S[ii + 1, jj] / denom
                            borrowing[ii, jj].append([ii + 1, jj, patch, None])
                            flag_intr_cell[ii + 1, jj] = True
                            S_enl[ii, jj] += patch
                            S_red[ii + 1, jj] -= patch
                        if local_avail[ii, jj - 1]:
                            patch = S_ext * S[ii, jj - 1] / denom
                            borrowing[ii, jj].append([ii, jj - 1, patch, None])
                            flag_intr_cell[ii, jj - 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii, jj - 1] -= patch
                        if local_avail[ii, jj + 1]:
                            patch = S_ext * S[ii, jj + 1] / denom
                            borrowing[ii, jj].append([ii, jj + 1, patch, None])
                            flag_intr_cell[ii, jj + 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii, jj + 1] -= patch
                        if local_avail[ii - 1, jj - 1]:
                            patch = S_ext * S[ii - 1, jj - 1] / denom
                            borrowing[ii, jj].append([ii - 1, jj - 1, patch, None])
                            flag_intr_cell[ii - 1, jj - 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii - 1, jj - 1] -= patch
                        if local_avail[ii + 1, jj - 1]:
                            patch = S_ext * S[ii + 1, jj - 1] / denom
                            borrowing[ii, jj].append([ii + 1, jj - 1, patch, None])
                            flag_intr_cell[ii + 1, jj - 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii + 1, jj - 1] -= patch
                        if local_avail[ii - 1, jj + 1]:
                            patch = S_ext * S[ii - 1, jj + 1] / denom
                            borrowing[ii, jj].append([ii - 1, jj + 1, patch, None])
                            flag_intr_cell[ii - 1, jj + 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii - 1, jj + 1] -= patch
                        if local_avail[ii + 1, jj + 1]:
                            patch = S_ext * S[ii + 1, jj + 1] / denom
                            borrowing[ii, jj].append([ii + 1, jj + 1, patch, None])
                            flag_intr_cell[ii + 1, jj + 1] = True
                            S_enl[ii, jj] += patch
                            S_red[ii + 1, jj + 1] -= patch

                        flag_ext_cell[ii, jj] = False


"""
Function to compute the area of the cells of the conformal grid.
"""


@jit("(f8[:,:], f8[:,:], f8[:,:], f8[:,:], i4, i4, f8, f8)", nopython=True)
def compute_areas(l_x, l_y, S, S_red, nx, ny, dx, dy):
    # Normalize the grid lengths for robustness
    l_x /= dx
    l_y /= dy

    # Loop over the cells
    for ii in range(nx):
        for jj in range(ny):
            # If at least 3 edges have full length consider the area as full
            # (this also takes care of the case in which an edge lies exactly on the boundary)
            # count_cane = eq(l_x[ii, jj], 1.0) + eq(l_y[ii, jj], 1.0) + eq(l_x[ii, jj + 1], 1.0) + eq(l_y[ii + 1, jj], 1.0)
            if (
                np.sum(
                    np.array(
                        [
                            eq(l_x[ii, jj], 1.0),
                            eq(l_y[ii, jj], 1.0),
                            eq(l_x[ii, jj + 1], 1.0),
                            eq(l_y[ii + 1, jj], 1.0),
                        ]
                    )
                )
                >= 3
            ):
                # if count_cane >=3:
                S[ii, jj] = 1.0
            elif (
                eq(l_x[ii, jj], 0.0)
                and neq(l_y[ii, jj], 0.0)
                and neq(l_x[ii, jj + 1], 0.0)
                and neq(l_y[ii + 1, jj], 0)
            ):
                S[ii, jj] = 0.5 * (l_y[ii, jj] + l_y[ii + 1, jj]) * l_x[ii, jj + 1]
            elif (
                eq(l_x[ii, jj + 1], 0.0)
                and neq(l_y[ii, jj], 0.0)
                and neq(l_x[ii, jj], 0.0)
                and neq(l_y[ii + 1, jj], 0.0)
            ):
                S[ii, jj] = 0.5 * (l_y[ii, jj] + l_y[ii + 1, jj]) * l_x[ii, jj]
            elif (
                eq(l_y[ii, jj], 0.0)
                and neq(l_x[ii, jj], 0.0)
                and neq(l_x[ii, jj + 1], 0.0)
                and neq(l_y[ii + 1, jj], 0.0)
            ):
                S[ii, jj] = 0.5 * (l_x[ii, jj] + l_x[ii, jj + 1]) * l_y[ii + 1, jj]
            elif (
                eq(l_y[ii + 1, jj], 0.0)
                and neq(l_x[ii, jj], 0.0)
                and neq(l_y[ii, jj], 0.0)
                and neq(l_x[ii, jj + 1], 0.0)
            ):
                S[ii, jj] = 0.5 * (l_x[ii, jj] + l_x[ii, jj + 1]) * l_y[ii, jj]
            elif (
                eq(l_x[ii, jj], 0.0)
                and eq(l_y[ii, jj], 0.0)
                and neq(l_x[ii, jj + 1], 0.0)
                and neq(l_y[ii + 1, jj], 0.0)
            ):
                S[ii, jj] = 0.5 * l_x[ii, jj + 1] * l_y[ii + 1, jj]
            elif (
                eq(l_x[ii, jj], 0.0)
                and eq(l_y[ii + 1, jj], 0.0)
                and neq(l_x[ii, jj + 1], 0.0)
                and neq(l_y[ii, jj], 0.0)
            ):
                S[ii, jj] = 0.5 * l_x[ii, jj + 1] * l_y[ii, jj]
            elif (
                eq(l_x[ii, jj + 1], 0.0)
                and eq(l_y[ii + 1, jj], 0.0)
                and neq(l_x[ii, jj], 0.0)
                and neq(l_y[ii, jj], 0.0)
            ):
                S[ii, jj] = 0.5 * l_x[ii, jj] * l_y[ii, jj]
            elif (
                eq(l_x[ii, jj + 1], 0.0)
                and eq(l_y[ii, jj], 0.0)
                and neq(l_x[ii, jj], 0.0)
                and neq(l_y[ii + 1, jj], 0.0)
            ):
                S[ii, jj] = 0.5 * l_x[ii, jj] * l_y[ii + 1, jj]
            elif (
                0.0 < l_x[ii, jj] <= 1.0
                and 0.0 < l_y[ii, jj] <= 1.0
                and eq(l_x[ii, jj + 1], 1.0)
                and eq(l_y[ii + 1, jj], 1.0)
            ):
                S[ii, jj] = 1.0 - 0.5 * (1.0 - l_x[ii, jj]) * (1.0 - l_y[ii, jj])
            elif (
                0.0 < l_x[ii, jj] <= 1.0
                and 0.0 < l_y[ii + 1, jj] <= 1.0
                and eq(l_x[ii, jj + 1], 1.0)
                and eq(l_y[ii, jj], 1.0)
            ):
                S[ii, jj] = 1.0 - 0.5 * (1.0 - l_x[ii, jj]) * (1.0 - l_y[ii + 1, jj])
            elif (
                0.0 < l_x[ii, jj + 1] <= 1.0
                and 0.0 < l_y[ii + 1, jj] <= 1.0
                and eq(l_x[ii, jj], 1.0)
                and eq(l_y[ii, jj], 1.0)
            ):
                S[ii, jj] = 1.0 - 0.5 * (1.0 - l_x[ii, jj + 1]) * (
                    1.0 - l_y[ii + 1, jj]
                )
            elif (
                0.0 < l_x[ii, jj + 1] <= 1.0
                and 0.0 < l_y[ii, jj] <= 1.0
                and eq(l_x[ii, jj], 1.0)
                and eq(l_y[ii + 1, jj], 1.0)
            ):
                S[ii, jj] = 1.0 - 0.5 * (1.0 - l_x[ii, jj + 1]) * (1.0 - l_y[ii, jj])

    l_x *= dx
    l_y *= dy


"""
Function to mark wich cells are interior (int), require extension (unst),
are on the boundary(bound), are available for intrusion (avail)
"""


@jit(
    "(f8[:,:], f8[:,:], i4, i4, f8, f8, f8[:,:], b1[:,:], f8[:,:], b1[:,:], b1[:,:], b1[:,:])",
    nopython=True,
)
def mark_cells(
    l_x,
    l_y,
    nx,
    ny,
    dx,
    dy,
    S,
    flag_int_cell,
    S_stab,
    flag_unst_cell,
    flag_bound_cell,
    flag_avail_cell,
):
    for ii in range(nx):
        for jj in range(ny):
            flag_int_cell[ii, jj] = S[ii, jj] > 0
            if flag_int_cell[ii, jj]:
                S_stab[ii, jj] = 0.5 * np.max(
                    np.array(
                        [
                            l_x[ii, jj] * dy,
                            l_x[ii, jj + 1] * dy,
                            l_y[ii, jj] * dx,
                            l_y[ii + 1, jj] * dx,
                        ]
                    )
                )
            flag_unst_cell[ii, jj] = S[ii, jj] < S_stab[ii, jj]
            flag_bound_cell[ii, jj] = (
                (0 < l_x[ii, jj] < dx)
                or (0 < l_y[ii, jj] < dy)
                or (0 < l_x[ii, jj + 1] < dx)
                or (0 < l_y[ii + 1, jj] < dy)
            )
            flag_avail_cell[ii, jj] = flag_int_cell[ii, jj] and (
                not flag_unst_cell[ii, jj]
            )



import numpy as np


class ConductorsAssembly:
    def __init__(self, conductors):
        self.conductors = conductors

    def in_conductor(self, x, y, z):
        for conductor in self.conductors:
            if conductor.in_conductor(x, y, z):
                return True

        return False

    def out_conductor(self, x, y, z):
        return not self.in_conductor(x, y, z)

    def intersec_x(self, x, y, z):
        list_inters = np.zeros_like(self.conductors)
        for ii, conductor in enumerate(self.conductors):
            list_inters[ii] = conductor.intersec_x(x, y, z)

        dist_inters = abs(list_inters - x)
        return list_inters[np.argmin(dist_inters)]

    def intersec_y(self, x, y, z):
        list_inters = np.zeros_like(self.conductors)
        for ii, conductor in enumerate(self.conductors):
            list_inters[ii] = conductor.intersec_y(x, y, z)

        dist_inters = abs(list_inters - y)
        return list_inters[np.argmin(dist_inters)]

    def intersec_z(self, x, y, z):
        list_inters = np.zeros_like(self.conductors)
        for ii, conductor in enumerate(self.conductors):
            list_inters[ii] = conductor.intersec_z(x, y, z)

        dist_inters = abs(list_inters - z)
        return list_inters[np.argmin(dist_inters)]


class InCube:
    def __init__(self, lx, ly, lz, x_cent, y_cent, z_cent):
        self.lx = lx
        self.ly = ly
        self.lz = lz
        self.x_cent = x_cent
        self.y_cent = y_cent
        self.z_cent = z_cent

    def out_conductor(self, x, y, z):
        return (
            -0.5 * self.lx + self.x_cent < x < 0.5 * self.lx + self.x_cent
            and -0.5 * self.ly + self.y_cent < y < 0.5 * self.ly + self.y_cent
            and -0.5 * self.lz + self.z_cent < z < 0.5 * self.lz + self.z_cent
        )

    def in_conductor(self, x, y, z):
        return not self.out_conductor(x, y, z)

    def intersec_x(self, x, y, z):
        inters_1 = -0.5 * self.lx + self.x_cent
        inters_2 = inters_1 + self.lx
        if abs(x - inters_1) < abs(x - inters_2):
            return inters_1
        else:
            return inters_2

    def intersec_y(self, x, y, z):
        inters_1 = -0.5 * self.ly + self.y_cent
        inters_2 = inters_1 + self.ly
        if abs(y - inters_1) < abs(y - inters_2):
            return inters_1
        else:
            return inters_2

    def intersec_z(self, x, y, z):
        inters_1 = -0.5 * self.lz + self.z_cent
        inters_2 = inters_1 + self.lz
        if abs(z - inters_1) < abs(z - inters_2):
            return inters_1
        else:
            return inters_2


class InSphere:
    def __init__(self, radius, x_cent, y_cent, z_cent):
        self.radius = radius
        self.x_cent = x_cent
        self.y_cent = y_cent
        self.z_cent = z_cent

    def in_conductor(self, x, y, z):
        return np.square(x - self.x_cent) + np.square(y - self.y_cent) + np.square(
            z - self.z_cent
        ) <= np.square(self.radius)

    def out_conductor(self, x, y, z):
        return not self.out_conductor(x, y, z)

    def intersec_x(self, x, y, z):
        inters_1 = (
            np.sqrt(
                np.square(self.radius)
                - np.square(y - self.y_cent)
                - np.square(z - self.z_cent)
            )
            + self.x_cent
        )
        inters_2 = -(
            np.sqrt(
                np.square(self.radius)
                - np.square(y - self.y_cent)
                - np.square(z - self.z_cent)
            )
            + self.x_cent
        )

        if abs(x - inters_1) < abs(x - inters_2):
            return inters_1
        else:
            return inters_2

    def intersec_y(self, x, y, z):
        inters_1 = (
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(z - self.z_cent)
            )
            + self.y_cent
        )
        inters_2 = -(
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(z - self.z_cent)
            )
            + self.y_cent
        )
        if abs(y - inters_1) < abs(y - inters_2):
            return inters_1
        else:
            return inters_2

    def intersec_z(self, x, y, z):
        inters_1 = (
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(y - self.y_cent)
            )
            + self.z_cent
        )
        inters_2 = -(
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(y - self.y_cent)
            )
            + self.z_cent
        )
        if abs(z - inters_1) < abs(z - inters_2):
            return inters_1
        else:
            return inters_2


class noConductor:
    def out_conductor(self, x, y, z):
        return True

    def in_conductor(self, x, y, z):
        return False

    def intersec_x(self, x, y, z):
        return 1000

    def intersec_y(self, x, y, z):
        return 1000

    def intersec_z(self, x, y, z):
        return 1000


class Plane:
    def __init__(self, p, n):
        self.p = p
        self.n = n

    def in_conductor(self, x, y, z):
        return (
            self.n[0] * (x - self.p[0])
            + self.n[1] * (y - self.p[1])
            + self.n[2] * (z - self.p[2])
            <= 0
        )

    def out_conductor(self, x, y, z):
        return not self.in_conductor(x, y, z)

    def intersec_x(self, x, y, z):
        if self.n[0] == 0:
            return 1000
        else:
            return (
                -(self.n[1] * (y - self.p[1]) + self.n[2] * (z - self.p[2])) / self.n[0]
                + self.p[0]
            )

    def intersec_y(self, x, y, z):
        if self.n[1] == 0:
            return 1000
        else:
            return (
                -(self.n[0] * (x - self.p[0]) + self.n[2] * (z - self.p[2])) / self.n[1]
                + self.p[1]
            )

    def intersec_z(self, x, y, z):
        if self.n[2] == 0:
            return 1000
        else:
            return (
                -(self.n[0] * (x - self.p[0]) + self.n[1] * (y - self.p[1])) / self.n[2]
                + self.p[2]
            )


class OutSphere:
    def __init__(self, radius, x_cent=0, y_cent=0, z_cent=0):
        self.radius = radius
        self.x_cent = x_cent
        self.y_cent = y_cent
        self.z_cent = z_cent

    def in_conductor(self, x, y, z):
        return np.square(x - self.x_cent) + np.square(y - self.y_cent) + np.square(
            z - self.z_cent
        ) >= np.square(self.radius)

    def out_conductor(self, x, y, z):
        return not self.in_conductor(x, y, z)

    def intersec_x(self, x, y, z):
        inters_1 = (
            np.sqrt(
                np.square(self.radius)
                - np.square(y - self.y_cent)
                - np.square(z - self.z_cent)
            )
            + self.x_cent
        )
        inters_2 = -(
            np.sqrt(
                np.square(self.radius)
                - np.square(y - self.y_cent)
                - np.square(z - self.z_cent)
            )
            + self.x_cent
        )

        if abs(x - inters_1) < abs(x - inters_2):
            return inters_1
        else:
            return inters_2

    def intersec_y(self, x, y, z):
        inters_1 = (
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(z - self.z_cent)
            )
            + self.y_cent
        )
        inters_2 = -(
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(z - self.z_cent)
            )
            + self.y_cent
        )
        if abs(y - inters_1) < abs(y - inters_2):
            return inters_1
        else:
            return inters_2

    def intersec_z(self, x, y, z):
        inters_1 = (
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(y - self.y_cent)
            )
            + self.z_cent
        )
        inters_2 = -(
            np.sqrt(
                np.square(self.radius)
                - np.square(x - self.x_cent)
                - np.square(y - self.y_cent)
            )
            + self.z_cent
        )
        if abs(z - inters_1) < abs(z - inters_2):
            return inters_1
        else:
            return inters_2



import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv

sys.path.append("../wakis")

import pytest
from scipy.constants import c
from tqdm import tqdm

from wakis import GridFIT3D, SolverFIT3D, WakeSolver
from wakis.sources import Beam

# Run with:
# mpiexec -n 2 python -m pytest --color=yes -v -s tests/test_007_mpi_lossy_cavity.py


@pytest.mark.slow
class TestMPILossyCavity:
    # Regression data
    # fmt: off
    tol = dict(rtol=50e-5, atol=50e-4)
    dtype = np.float32

    WP = np.array([-1.20513623e-18 ,-3.43161145e-14 ,-9.12255548e-11 ,-5.96997512e-08,
                    -1.14173019e-05 ,-6.67499094e-04 ,-1.20239747e-02 ,-6.53062215e-02,
                    -9.26520639e-02 , 2.78321623e-02  ,1.25518516e-01 , 7.30670041e-02,
                    -1.90239810e-02 ,-4.85596408e-02 ,-3.20357619e-02 ,-1.41604640e-02,
                    6.08356174e-02 , 8.46570691e-02 ,-5.47181776e-02 ,-1.16180823e-01,
                    1.96111216e-02 , 9.92679050e-02 , 2.71878152e-02 ,-5.20035156e-02,
                    -4.06137472e-02 ,-1.13185667e-02 , 1.23530805e-02 , 7.68524041e-02,
                    2.25859758e-02 ,-9.37436643e-02 ,-5.75287185e-02 , 6.39890431e-02,
                    8.01280003e-02 ,-2.25671230e-02 ,-5.80154107e-02 ,-1.67525750e-02,
                    1.87720028e-03 , 4.10247646e-02  ,5.36731566e-02 ,-3.30682943e-02,
                    -8.35234400e-02 ,-3.49965114e-03 , 8.57976513e-02 , 3.52580927e-02,
                    -5.59291793e-02 ,-3.74168173e-02 ,-1.57790071e-03 , 2.16477574e-02,
                    4.65687493e-02 , 1.13356268e-02 ,-5.61946649e-02 ,-5.45040228e-02,
                    4.41993189e-02  ,7.38920452e-02 ,-1.60854441e-02 ,-5.54896162e-02,
                    -1.77145556e-02 , 1.61884877e-02 , 3.33170207e-02 , 2.63165394e-02,
                    -1.83733826e-02 ,-5.94726314e-02 ,-1.15493543e-02 , 6.90379083e-02,
                    3.30361171e-02 ,-4.32087737e-02 ,-4.14862834e-02 , 2.95742481e-03,
                    3.11653273e-02 , 2.43513549e-02 , 5.17266702e-03 ,-3.70591561e-02,
                    -4.24818163e-02 , 3.09681557e-02 , 5.76902933e-02 ,-4.78519848e-03,
                    -5.02625318e-02 ,-2.09821868e-02 , 2.46893459e-02 , 2.77009485e-02,
                    1.12395692e-02 ,-1.34117335e-02 ,-4.22375110e-02 ,-8.34037808e-03,
                    4.81709861e-02 , 3.05819110e-02 ,-2.98222467e-02 ,-4.38079436e-02,
                    6.51694708e-03 , 3.11637911e-02 , 1.61475003e-02 ,-1.03231050e-03,
                    -2.84581274e-02 ,-2.64695160e-02  ,1.92431962e-02 , 4.34036616e-02,
                    3.00897180e-03 ,-4.36364833e-02 ,-2.09601909e-02 , 2.48741294e-02,
                    2.47321674e-02  ,5.34334063e-03 ,-1.56819510e-02 ,-2.75503078e-02])

    Z = np.array([ 7.77970188e+00   -0.j      ,   -3.85415882e+00  +14.7587971j,
                -3.61177110e+00   +4.52999754j,  1.11502695e+01  +16.94389961j,
                -1.48367615e+00  +31.13414676j ,-2.42369181e+00  +20.20481814j,
                1.32631009e+01  +32.26388931j , 7.84469425e-02  +48.06727516j,
                -2.04548300e+00  +36.16821013j , 1.55741147e+01  +48.44883647j,
                1.37311998e+00  +66.93221773j ,-2.19747672e+00  +53.47374311j,
                1.86245241e+01  +66.56951352j  ,2.65234275e+00  +89.33983678j,
                -2.94989399e+00  +73.40544312j , 2.31553338e+01  +88.22587886j,
                4.16269453e+00 +118.22389548j ,-4.66524583e+00  +98.31868989j,
                3.07626943e+01 +116.67578417j , 6.41211027e+00 +160.3569522j,
                -8.36053398e+00 +133.90634108j , 4.61970915e+01 +160.6857645j,
                1.14408149e+01 +237.02472112j ,-1.72907188e+01 +200.58798515j,
                9.34171798e+01 +257.88692816j , 4.19201258e+01 +475.3131594j,
                -3.28372915e+01 +498.40274299j , 1.47161715e+03 +896.91077291j,
                4.46726052e+02-1074.75272128j , 1.48368867e+02 -150.95840903j,
                -7.65707535e+01  -95.4006897j ,  4.13011813e+01 -115.5293095j,
                5.96957058e+01  +64.38392007j ,-4.17351813e+01  +65.90769948j,
                4.16431845e+01  +45.42941094j , 5.22637205e+01 +182.75009977j,
                -2.42252500e+01 +183.45582422j , 8.56170191e+01 +206.784262j,
                9.77958353e+01 +444.21692899j , 8.26325734e+01 +560.54204926j,
                1.44414393e+03 +765.73788438j , 7.29867370e+02-1096.11716581j,
                1.17986711e+02 -317.31632024j ,-4.56173056e+01 -277.3675436j,
                7.54525872e+01 -200.5232469j ,  2.20725250e+01  -36.10606952j,
                -3.94369541e+01  -81.45040661j , 6.08389766e+01  -53.3224972j,
                1.97319998e+01  +64.35491532j ,-1.82274232e+01   +4.8794245j])

    Ez = np.array([ 2.21799580e+01, -5.78362380e+00 ,-6.00531608e+00 ,-1.01718118e+00,
                    1.73468186e+00 , 1.29313000e+00 ,-9.69312994e+00 ,-2.59107426e+01,
                    -4.73006858e+01 ,-7.11016384e+01 ,-9.36971115e+01 ,-1.12676454e+02,
                    -1.28296571e+02 ,-1.40866374e+02 ,-1.48363099e+02 ,-1.51635428e+02,
                    -1.48185125e+02 ,-1.37740237e+02 ,-1.19948398e+02 ,-9.23354987e+01,
                    -6.04105997e+01 ,-3.06532160e+01 ,-1.17749936e+01 ,-3.12574866e+00,
                    -7.35339521e-01 ,-1.13085658e-01 , 7.18247535e-01 , 8.73829036e-02])

    # fmt: on

    gridLogs = {
        "use_mesh_refinement": False,
        "Nx": 60,
        "Ny": 60,
        "Nz": 140,
        "dx": 0.00866666634877522,
        "dy": 0.00866666634877522,
        "dz": 0.005714285799435207,
        "xmin": -0.25999999046325684,
        "xmax": 0.25999999046325684,
        "ymin": -0.25999999046325684,
        "ymax": 0.25999999046325684,
        "zmin": -0.25,
        "zmax": 0.550000011920929,
        "stl_solids": {
            "cavity": "tests/stl/007_vacuum_cavity.stl",
            "shell": "tests/stl/007_lossymetal_shell.stl",
        },
        "stl_materials": {"cavity": "vacuum", "shell": [30, 1.0, 30]},
        "gridInitializationTime": 0,
    }

    solverLogs = {
        "use_gpu": False,  # updated in test_log_file
        "use_mpi": False,  # updated in test_log_file
        "background": "pec",
        "bc_low": ["pec", "pec", "pec"],
        "bc_high": ["pec", "pec", "pec"],
        "dt": dtype(6.970326659611059e-12),
        "solverInitializationTime": 0,
    }

    wakeSolverLogs = {
        "ti": 2.8516132094735135e-09,
        "q": 1e-09,
        "sigmaz": 0.1,
        "beta": 1.0,
        "xsource": 0.0,
        "ysource": 0.0,
        "xtest": 0.0,
        "ytest": 0.0,
        "chargedist": None,
        "skip_cells": 10,
        "results_folder": "tests/007_results/",
        "wakelength": 10.0,
        "simulationTime": 0,
    }

    img_folder = "tests/007_img/"

    def test_mpi_import(self):
        # ---------- MPI setup ------------
        global use_mpi
        try:
            # can be skipped since it is handled inside GridFIT3D
            from mpi4py import MPI

            comm = MPI.COMM_WORLD  # Get MPI communicator
            size = comm.Get_size()  # Total number of MPI processes
            if size > 1:
                use_mpi = True
            else:
                use_mpi = False
        except Exception as e:
            print(f"[!] MPI not available: {e}")
            use_mpi = False

        print(f"Using mpi: {use_mpi}")

    def test_mpi_simulation(self, use_gpu):
        # ---------- Domain setup ---------

        # Geometry & Materials
        solid_1 = "tests/stl/007_vacuum_cavity.stl"
        solid_2 = "tests/stl/007_lossymetal_shell.stl"

        stl_solids = {"cavity": solid_1, "shell": solid_2}

        stl_materials = {
            "cavity": "vacuum",
            "shell": [30, 1.0, 30],  # [eps_r, mu_r, sigma[S/m]]
        }

        # Extract domain bounds from geometry
        solids = pv.read(solid_1) + pv.read(solid_2)
        xmin, xmax, ymin, ymax, ZMIN, ZMAX = solids.bounds

        # Number of mesh cells
        Nx = 60
        Ny = 60
        NZ = 140
        global use_mpi
        grid = GridFIT3D(
            xmin,
            xmax,
            ymin,
            ymax,
            ZMIN,  # Global domain zmin
            ZMAX,  # Global domain zmax
            Nx,
            Ny,
            NZ,  # Global domain Nz
            use_mpi=use_mpi,  # Enables MPI subdivision of the domain
            stl_solids=stl_solids,
            stl_materials=stl_materials,
            stl_scale=1.0,
            stl_rotate=[0, 0, 0],
            stl_translate=[0, 0, 0],
            verbose=1,
        )
        if use_mpi:
            print(
                f"Process {grid.rank}: Handling Z range {grid.zmin} to {grid.zmax} with {grid.Nz} cells"
            )

        # ------------ Beam source & Wake ----------------
        # Beam parameters
        sigmaz = 10e-2  # [m] -> 2 GHz
        q = 1e-9  # [C]
        beta = 1.0  # beam beta
        xs = 0.0  # x source position [m]
        ys = 0.0  # y source position [m]
        ti = 3 * sigmaz / c  # injection time [s]

        beam = Beam(
            q=q, sigmaz=sigmaz, beta=beta, xsource=xs, ysource=ys, ti=ti
        )

        # ----------- Solver & Simulation ----------
        # boundary conditions
        bc_low = ["pec", "pec", "pec"]
        bc_high = ["pec", "pec", "pec"]

        # Solver setup
        global solver
        solver = SolverFIT3D(
            grid,
            bc_low=bc_low,
            bc_high=bc_high,
            use_stl=True,
            use_mpi=use_mpi,  # Activate MPI
            bg="pec",  # Background material
            dtype=self.dtype,
            use_gpu=use_gpu,
        )

        # -------------- Output folder ---------------------
        if use_mpi and solver.rank == 0:
            if not os.path.exists(self.img_folder):
                os.mkdir(self.img_folder)
        elif not use_mpi:
            if not os.path.exists(self.img_folder):
                os.mkdir(self.img_folder)

        # -------------- Custom time loop  -----------------
        if use_mpi:
            Nt = 3000
            for n in tqdm(range(Nt)):
                beam.update(solver, n * solver.dt)
                solver.one_step()  # MPI handled internally

            Ez = solver.mpi_gather("Ez", x=int(Nx / 2), y=int(Ny / 2))
            if solver.rank == 0:
                # print(Ez)
                # print(len(Ez))
                assert len(Ez) == NZ, (
                    "Electric field Ez samples length mismatch"
                )
                assert np.allclose(Ez[np.s_[::5]], self.Ez, **self.tol), (
                    "Electric field Ez samples MPI failed"
                )
        else:
            Nt = 3000
            for n in tqdm(range(Nt)):
                beam.update(solver, n * solver.dt)
                solver.one_step()

            Ez = solver.E[int(Nx / 2), int(Ny / 2), np.s_[::5], "z"]
            # print(Ez)
            assert len(solver.E[int(Nx / 2), int(Ny / 2), :, "z"]) == NZ, (
                "Electric field Ez samples length mismatch"
            )
            assert np.allclose(Ez, self.Ez, **self.tol), (
                "Electric field Ez samples failed"
            )

    def test_mpi_gather_asField(self, flag_offscreen):
        # Plot inspect after mpi gather
        global solver
        if use_mpi:
            E = solver.mpi_gather_asField("E")
            if solver.rank == 0:  # custom plots go in rank 0
                fig, ax = E.inspect(
                    figsize=[20, 6],
                    plane="YZ",
                    off_screen=flag_offscreen,
                    handles=True,
                )
                fig.savefig(
                    self.img_folder + "Einspect_" + str(3000).zfill(4) + ".png"
                )
                plt.close(fig)
        else:
            fig, ax = solver.E.inspect(
                figsize=[20, 6],
                plane="YZ",
                off_screen=flag_offscreen,
                handles=True,
            )
            fig.savefig(
                self.img_folder + "Einspect_" + str(3000).zfill(4) + ".png"
            )
            plt.close(fig)

    def test_mpi_plot2D(self, flag_offscreen):
        # Plot E abs in 2D every 20 timesteps
        global solver
        solver.plot2D(
            field="E",
            component="Abs",
            plane="YZ",
            pos=0.5,
            cmap="rainbow",
            vmin=0,
            vmax=500.0,
            interpolation="hanning",
            off_screen=flag_offscreen,
            title=self.img_folder + "Ez2d",
            n=3000,
        )

    def test_mpi_plot1D(self, flag_offscreen):
        # Plot E z in 1D at diferent transverse positions `pos` every 20 timesteps
        global solver
        solver.plot1D(
            field="E",
            component="z",
            line="z",
            pos=[0.45, 0.5, 0.55],
            xscale="linear",
            yscale="linear",
            off_screen=flag_offscreen,
            title=self.img_folder + "Ez1d",
            n=3000,
        )

    def test_mpi_wakefield(self, use_gpu):
        # Reset fields
        global solver
        solver.reset_fields()

        # ------------ Beam source ----------------
        # Beam parameters
        sigmaz = 10e-2  # [m] -> 2 GHz
        q = 1e-9  # [C]
        beta = 1.0  # beam beta
        xs = 0.0  # x source position [m]
        ys = 0.0  # y source position [m]
        xt = 0.0  # x test position [m]
        yt = 0.0  # y test position [m]
        # [DEFAULT] tinj = 8.53*sigmaz/c_light  # injection time offset [s]

        # ----------- Wake Solver  setup  ----------
        # Wakefield post-processor
        wakelength = 10.0  # [m] -> Partially decayed
        skip_cells = 10  # no. cells to skip at zlo/zhi for wake integration
        results_folder = "tests/007_results/"

        global wake
        wake = WakeSolver(
            q=q,
            sigmaz=sigmaz,
            beta=beta,
            xsource=xs,
            ysource=ys,
            xtest=xt,
            ytest=yt,
            skip_cells=skip_cells,
            results_folder=results_folder,
            Ez_file=results_folder + "Ez.h5",
        )

        # Run simulation
        solver.wakesolve(wakelength=wakelength, wake=wake)

    def test_long_wake_potential(self):
        global wake
        global solver
        if use_mpi:
            if solver.rank == 0:
                tol = dict(rtol=0.1)
                assert len(wake.WP) == 5195, (
                    "Wake potential MPI samples length mismatch"
                )
                assert np.allclose(wake.WP[::50], self.WP, **tol), (
                    "Wake potential MPI samples failed"
                )
                assert np.cumsum(np.abs(wake.WP))[-1] == pytest.approx(
                    184.43818552913254, 0.1
                ), "Wake potential cumsum MPI failed"
        else:
            assert len(wake.WP) == 5195, (
                "Wake potential samples length mismatch"
            )
            assert np.allclose(wake.WP[::50], self.WP, **self.tol), (
                "Wake potential samples failed"
            )
            assert np.cumsum(np.abs(wake.WP))[-1] == pytest.approx(
                184.43818552913254, 0.1
            ), "Wake potential cumsum MPI failed"

    def test_long_impedance(self):
        global wake
        global solver
        if use_mpi:
            if solver.rank == 0:
                tol = dict(rtol=0.1)
                assert len(wake.Z) == 998, "Impedance samples length mismatch"
                assert np.allclose(
                    np.abs(wake.Z)[::20], np.abs(self.Z), **tol
                ), "Abs Impedance samples MPI failed"
                assert np.allclose(
                    np.real(wake.Z)[::20], np.real(self.Z), **tol
                ), "Real Impedance samples MPI failed"
                assert np.allclose(
                    np.imag(wake.Z)[::20], np.imag(self.Z), **tol
                ), "Imag Impedance samples MPI failed"
                assert np.cumsum(np.abs(wake.Z))[-1] == pytest.approx(
                    250910.51090497518, 0.1
                ), "Abs Impedance cumsum MPI failed"
        else:
            # print(wake.Z[::20])
            assert len(wake.Z) == 998, "Impedance samples length mismatch"
            assert np.allclose(
                np.abs(wake.Z)[::20], np.abs(self.Z), **self.tol
            ), "Abs Impedance samples failed"
            assert np.allclose(
                np.real(wake.Z)[::20], np.real(self.Z), **self.tol
            ), "Real Impedance samples failed"
            assert np.allclose(
                np.imag(wake.Z)[::20], np.imag(self.Z), **self.tol
            ), "Imag Impedance samples failed"
            assert np.cumsum(np.abs(wake.Z))[-1] == pytest.approx(
                250910.51090497518, 0.1
            ), "Abs Impedance cumsum failed"

    def test_log_file(self, use_gpu):
        # Helper function to compare nested dicts with float tolerance
        def assert_dict_allclose(d1, d2, rtol=1e-6, atol=1e-6, path=""):
            assert set(d1.keys()) == set(d2.keys()), (
                f"Key mismatch at {path}: {set(d1.keys())} != {set(d2.keys())}"
            )

            for k in d1:
                v1, v2 = d1[k], d2[k]
                p = f"{path}.{k}" if path else k

                # nested dict
                if isinstance(v1, dict) and isinstance(v2, dict):
                    assert_dict_allclose(v1, v2, rtol, atol, p)

                # floats
                elif isinstance(v1, float) and isinstance(v2, float):
                    if k == "dt":
                        assert v1 <= v2, (
                            "Timestep bigger than for uniform grid"
                        )
                    else:
                        assert np.isclose(v1, v2, rtol=rtol, atol=atol), (
                            f"Float mismatch at {p}: {v1} != {v2}"
                        )

                # np.floats
                elif isinstance(v1, np.floating) and isinstance(
                    v2, np.floating
                ):
                    if k == "dt":
                        assert v1 <= v2, (
                            "Timestep bigger than for uniform grid"
                        )
                    else:
                        assert np.isclose(v1, v2, rtol=rtol, atol=atol), (
                            f"Float mismatch at {p}: {v1} != {v2}"
                        )

                # lists/tuples/arrays
                elif isinstance(v1, (list, tuple, np.ndarray)) and isinstance(
                    v2, (list, tuple, np.ndarray)
                ):
                    assert len(v1) == len(v2), f"Length mismatch at {p}"
                    for i, (a, b) in enumerate(zip(v1, v2)):
                        if isinstance(a, float) and isinstance(b, float):
                            assert np.isclose(a, b, rtol=rtol, atol=atol), (
                                f"Float mismatch at {p}[{i}]: {a} != {b}"
                            )
                        else:
                            assert a == b, (
                                f"Value mismatch at {p}[{i}]: {a} != {b}"
                            )

                # list vs single float
                elif isinstance(v1, (list, tuple, np.ndarray)) and isinstance(
                    v2, float
                ):
                    for i, a in enumerate(v1):
                        assert np.isclose(a, v2, rtol=rtol, atol=atol), (
                            f"Float mismatch at {p}[{i}]: {a} != {v2}"
                        )

                elif isinstance(v1, float) and isinstance(
                    v2, (list, tuple, np.ndarray)
                ):
                    for i, b in enumerate(v2):
                        assert np.isclose(v1, b, rtol=rtol, atol=atol), (
                            f"Float mismatch at {p}[{i}]: {v1} != {b}"
                        )

                # everything else → exact match
                else:
                    assert v1 == v2, f"Mismatch at {p}: {v1} != {v2}"

        global solver
        # Exclude timing info from comparison as they can vary between runs
        solver.logger.grid["gridInitializationTime"] = 0
        solver.logger.solver["solverInitializationTime"] = 0
        solver.logger.wakeSolver["simulationTime"] = 0
        self.solverLogs["use_mpi"] = use_mpi
        self.solverLogs["use_gpu"] = use_gpu

        # Check log file exists
        logfile = os.path.join(
            solver.logger.wakeSolver["results_folder"], "wakis.log"
        )
        assert os.path.exists(logfile), "Log file not created"

        # Compare log dict contents
        assert_dict_allclose(solver.logger.grid, self.gridLogs)
        assert_dict_allclose(solver.logger.solver, self.solverLogs)
        assert_dict_allclose(solver.logger.wakeSolver, self.wakeSolverLogs)



import numpy as np
import scipy.optimize


class OutRect:
    def __init__(self, Lx, Ly, x_cent, y_cent):
        self.Lx = Lx
        self.Ly = Ly
        self.x_cent = x_cent
        self.y_cent = y_cent
        # self.theta = theta
        # self.R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
        # self.mR = np.array([[np.cos(-theta), -np.sin(-theta)], [np.sin(-theta), np.cos(-theta)]])

    def out_conductor(self, x, y):
        # [xx, yy] = np.dot(self.mR, np.array([x, y]))
        return (-0.5 * self.Lx + self.x_cent < x < 0.5 * self.Lx + self.x_cent) and (
            -0.5 * self.Ly + self.y_cent < y < 0.5 * self.Ly + self.y_cent
        )

    def in_conductor(self, x, y):
        return not self.out_conductor(x, y)

    def intersec_x(self, x, y):
        # [xx, yy] = np.dot(self.mR, np.array([x, y]))
        inters_1 = -0.5 * self.Lx + self.x_cent
        inters_2 = inters_1 + self.Lx
        if abs(x - inters_1) < abs(x - inters_2):
            return inters_1
        else:
            return inters_2

        # [xxx, _] = np.dot(self.R, np.array([inters, yy]))
        # return xxx

    def intersec_y(self, x, y):
        # [xx, yy] = np.dot(self.mR, np.array([x, y]))
        inters_1 = -0.5 * self.Ly + self.y_cent
        inters_2 = inters_1 + self.Ly
        if abs(y - inters_1) < abs(y - inters_2):
            return inters_1
        else:
            return inters_2

        # [_, yyy] = np.dot(self.R, np.array([xx, inters]))
        # return yyy


class ImpFunc:
    def __init__(self, func):
        self.func = func

    def out_conductor(self, x, y):
        return self.func(x, y) < 0

    def in_conductor(self, x, y):
        return self.func(x, y) > 0

    def intersec_x(self, x, y):
        func_x = lambda t: self.func(t, y)

        return scipy.optimize.newton_krylov(func_x, x)

    def intersec_y(self, x, y):
        func_y = lambda t: self.func(x, t)

        return scipy.optimize.newton_krylov(func_y, y)


class Plane:
    def __init__(self, m_plane, q_plane, tol=0, sign=1):
        self.tol = tol  # 1e-16
        self.m_plane = m_plane
        self.q_plane = q_plane
        self.sign = sign

    def in_conductor(self, x, y):
        if self.sign == 1:
            return y - self.m_plane * x - self.q_plane >= self.tol
        elif self.sign == -1:
            return y - self.m_plane * x - self.q_plane <= self.tol
        else:
            print("sign must be + or - 1")

    def out_conductor(self, x, y):
        return not self.in_conductor(x, y)

    def intersec_x(self, x, y):
        return y / self.m_plane - self.q_plane / self.m_plane

    def intersec_y(self, x, y):
        return self.m_plane * x + self.q_plane


class InCircle:
    def __init__(self, radius, x_cent, y_cent):
        self.radius = radius
        self.x_cent = x_cent
        self.y_cent = y_cent

    def in_conductor(self, x, y):
        return np.square(x - self.x_cent) + np.square(y - self.y_cent) <= np.square(
            self.radius
        )

    def out_conductor(self, x, y):
        return not self.in_conductor(x, y)

    def intersec_x(self, x, y):
        if abs(y - self.y_cent) <= self.radius:
            inters_1 = (
                np.sqrt(np.square(self.radius) - np.square(y - self.y_cent))
                + self.y_cent
            )
            inters_2 = (
                -np.sqrt(np.square(self.radius) - np.square(y - self.y_cent))
                + self.y_cent
            )
            if abs(x - inters_1) < abs(x - inters_2):
                return inters_1
            else:
                return inters_2
        else:
            return np.inf

    def intersec_y(self, x, y):
        if abs(x - self.x_cent) <= self.radius:
            inters_1 = (
                np.sqrt(np.square(self.radius) - np.square(x - self.x_cent))
                + self.x_cent
            )
            inters_2 = (
                -np.sqrt(np.square(self.radius) - np.square(x - self.x_cent))
                + self.x_cent
            )
            if abs(y - inters_1) < abs(y - inters_2):
                return inters_1
            else:
                return inters_2
        else:
            return np.inf


class OutCircle:
    def __init__(self, radius, x_cent, y_cent):
        self.radius = radius
        self.x_cent = x_cent
        self.y_cent = y_cent

    def in_conductor(self, x, y):
        return np.square(x - self.x_cent) + np.square(y - self.y_cent) >= np.square(
            self.radius
        )

    def out_conductor(self, x, y):
        return not self.in_conductor(x, y)

    def intersec_x(self, x, y):
        # if abs(y - self.y_cent) > self.radius:
        inters_1 = (
            np.sqrt(np.square(self.radius) - np.square(y - self.y_cent)) + self.x_cent
        )
        inters_2 = (
            -np.sqrt(np.square(self.radius) - np.square(y - self.y_cent)) + self.x_cent
        )
        if abs(x - inters_1) < abs(x - inters_2):
            return inters_1
        else:
            return inters_2
        # else:
        #    return np.inf

    def intersec_y(self, x, y):
        # if abs(x - self.x_cent) > self.radius:
        inters_1 = (
            np.sqrt(np.square(self.radius) - np.square(x - self.x_cent)) + self.y_cent
        )
        inters_2 = (
            -np.sqrt(np.square(self.radius) - np.square(x - self.x_cent)) + self.y_cent
        )
        if abs(y - inters_1) < abs(y - inters_2):
            return inters_1
        else:
            return inters_2
        # else:
        #    return np.inf


class ConductorsAssembly:
    def __init__(self, conductors):
        self.conductors = conductors

    def in_conductor(self, x, y):
        for conductor in self.conductors:
            if conductor.in_conductor(x, y):
                return True

        return False

    def out_conductor(self, x, y):
        return not self.in_conductor(x, y)

    def intersec_x(self, x, y):
        list_inters = np.zeros_like(self.conductors)
        for ii, conductor in enumerate(self.conductors):
            list_inters[ii] = conductor.intersec_x(x, y)

        dist_inters = abs(list_inters - x)
        return list_inters[np.argmin(dist_inters)]

    def intersec_y(self, x, y):
        list_inters = np.zeros_like(self.conductors)
        for ii, conductor in enumerate(self.conductors):
            list_inters[ii] = conductor.intersec_y(x, y)

        dist_inters = abs(list_inters - y)
        return list_inters[np.argmin(dist_inters)]


class noConductor:
    def out_conductor(self, x, y):
        return True

    def in_conductor(self, x, y):
        return False

    def intersec_x(self, x, y):
        return 1000

    def intersec_y(self, x, y):
        return 1000



import numpy as np
from grid2D import Grid2D
from grid2D import compute_areas as compute_areas_2D, mark_cells as mark_cells_2D
from numba import jit
from field import Field


def seg_length(x_1, y_1, z_1, x_2, y_2, z_2):
    return np.linalg.norm(np.array([x_1 - x_2, y_1 - y_2, z_1 - z_2]))


def eq(a, b, tol=1e-8):
    return abs(a - b) < tol


def neq(a, b, tol=1e-8):
    return not eq(a, b, tol)


class Grid3D:
    """
    Class holding the grid info and the routines for cell extensions.
      Constructor arguments:
          - xmin, xmax, ymin, ymax, zmin, zmax: extent of the domain.
          - nx, ny, nz: number of cells per direction
          - conductors: conductor object
          - sol_type: type of solver. 'FDTD' for staircased FDTD, 'DM' for Conformal Dey-Mittra FDTD,
                      'ECT' for Enlarged Cell Technique conformal FDTD
    """

    def __init__(
        self, xmin, xmax, ymin, ymax, zmin, zmax, nx, ny, nz, conductors, sol_type
    ):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.zmin = zmin
        self.zmax = zmax
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.dx = (xmax - xmin) / nx
        self.dy = (ymax - ymin) / ny
        self.dz = (ymax - ymin) / nz
        self.conductors = conductors

        self.l_x = np.zeros((nx, ny + 1, nz + 1))
        self.l_y = np.zeros((nx + 1, ny, nz + 1))
        self.l_z = np.zeros((nx + 1, ny + 1, nz))
        self.Sxy = np.zeros((nx, ny, nz + 1))
        self.Syz = np.zeros((nx + 1, ny, nz))
        self.Szx = np.zeros((nx, ny + 1, nz))
        self.Sxy_stab = np.zeros_like(self.Sxy)
        self.Sxy_enl = np.zeros_like(self.Sxy)
        self.Sxy_red = np.zeros_like(self.Sxy)
        self.Syz_stab = np.zeros_like(self.Syz)
        self.Syz_enl = np.zeros_like(self.Syz)
        self.Syz_red = np.zeros_like(self.Syz)
        self.Szx_stab = np.zeros_like(self.Szx)
        self.Szx_enl = np.zeros_like(self.Szx)
        self.Szx_red = np.zeros_like(self.Szx)
        self.flag_unst_cell_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.flag_intr_cell_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.flag_int_cell_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.flag_bound_cell_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.flag_avail_cell_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.flag_ext_cell_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.flag_unst_cell_yz = np.zeros_like(self.Syz, dtype=bool)
        self.flag_intr_cell_yz = np.zeros_like(self.Syz, dtype=bool)
        self.flag_int_cell_yz = np.zeros_like(self.Syz, dtype=bool)
        self.flag_bound_cell_yz = np.zeros_like(self.Syz, dtype=bool)
        self.flag_avail_cell_yz = np.zeros_like(self.Syz, dtype=bool)
        self.flag_ext_cell_yz = np.zeros_like(self.Syz, dtype=bool)
        self.flag_unst_cell_zx = np.zeros_like(self.Szx, dtype=bool)
        self.flag_intr_cell_zx = np.zeros_like(self.Szx, dtype=bool)
        self.flag_int_cell_zx = np.zeros_like(self.Szx, dtype=bool)
        self.flag_bound_cell_zx = np.zeros_like(self.Szx, dtype=bool)
        self.flag_avail_cell_zx = np.zeros_like(self.Szx, dtype=bool)
        self.flag_ext_cell_zx = np.zeros_like(self.Szx, dtype=bool)
        self.broken_xy = np.zeros_like(self.Sxy, dtype=bool)
        self.broken_yz = np.zeros_like(self.Syz, dtype=bool)
        self.broken_zx = np.zeros_like(self.Szx, dtype=bool)

        if (
            (sol_type != "FDTD")
            and (sol_type != "DM")
            and (sol_type != "ECT")
            and (sol_type != "FIT")
        ):
            raise ValueError(
                "sol_type must be:\n"
                + "\t'FDTD' for standard staircased FDTD\n"
                + "\t'DM' for Dey-Mittra conformal FDTD\n"
                + "\t'ECT' for Enlarged Cell Technique conformal FDTD"
            )

        self.compute_edges()
        if sol_type == "DM" or sol_type == "FDTD":  # or sol_type is 'FIT':
            self.compute_areas(
                self.l_x,
                self.l_y,
                self.l_z,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.Sxy_red,
                self.Syz_red,
                self.Szx_red,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
            )
            self.mark_cells(
                self.l_x,
                self.l_y,
                self.l_z,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.flag_int_cell_xy,
                self.flag_int_cell_yz,
                self.flag_int_cell_zx,
                self.Sxy_stab,
                self.Syz_stab,
                self.Szx_stab,
                self.flag_unst_cell_xy,
                self.flag_unst_cell_yz,
                self.flag_unst_cell_zx,
                self.flag_bound_cell_xy,
                self.flag_bound_cell_yz,
                self.flag_bound_cell_zx,
                self.flag_avail_cell_xy,
                self.flag_avail_cell_yz,
                self.flag_avail_cell_zx,
            )
        elif sol_type == "ECT":
            self.compute_areas(
                self.l_x,
                self.l_y,
                self.l_z,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.Sxy_red,
                self.Syz_red,
                self.Szx_red,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
            )
            self.mark_cells(
                self.l_x,
                self.l_y,
                self.l_z,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.flag_int_cell_xy,
                self.flag_int_cell_yz,
                self.flag_int_cell_zx,
                self.Sxy_stab,
                self.Syz_stab,
                self.Szx_stab,
                self.flag_unst_cell_xy,
                self.flag_unst_cell_yz,
                self.flag_unst_cell_zx,
                self.flag_bound_cell_xy,
                self.flag_bound_cell_yz,
                self.flag_bound_cell_zx,
                self.flag_avail_cell_xy,
                self.flag_avail_cell_yz,
                self.flag_avail_cell_zx,
            )

            # info about intruded cells (i,j,[(i_borrowing,j_borrowing,area_borrowing, )])
            self.borrowing_xy = np.empty((nx, ny, nz + 1), dtype=object)
            self.borrowing_yz = np.empty((nx + 1, ny, nz), dtype=object)
            self.borrowing_zx = np.empty((nx, ny + 1, nz), dtype=object)

            for ii in range(nx):
                for jj in range(ny):
                    for kk in range(nz + 1):
                        self.borrowing_xy[ii, jj, kk] = []
            for ii in range(nx + 1):
                for jj in range(ny):
                    for kk in range(nz):
                        self.borrowing_yz[ii, jj, kk] = []
            for ii in range(nx):
                for jj in range(ny + 1):
                    for kk in range(nz):
                        self.borrowing_zx[ii, jj, kk] = []

            self.flag_ext_cell_xy = self.flag_unst_cell_xy.copy()
            self.flag_ext_cell_yz = self.flag_unst_cell_yz.copy()
            self.flag_ext_cell_zx = self.flag_unst_cell_zx.copy()
            self.mark_cells(
                self.l_x,
                self.l_y,
                self.l_z,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.flag_int_cell_xy,
                self.flag_int_cell_yz,
                self.flag_int_cell_zx,
                self.Sxy_stab,
                self.Syz_stab,
                self.Szx_stab,
                self.flag_unst_cell_xy,
                self.flag_unst_cell_yz,
                self.flag_unst_cell_zx,
                self.flag_bound_cell_xy,
                self.flag_bound_cell_yz,
                self.flag_bound_cell_zx,
                self.flag_avail_cell_xy,
                self.flag_avail_cell_yz,
                self.flag_avail_cell_zx,
            )
            self.compute_extensions()

        elif sol_type == "FIT":
            # primal Grid G
            self.x = np.linspace(self.xmin, self.xmax, self.nx + 1)
            self.y = np.linspace(self.ymin, self.ymax, self.ny + 1)
            self.z = np.linspace(self.zmin, self.zmax, self.nz + 1)

            Y, X, Z = np.meshgrid(self.y, self.x, self.z)

            self.L = Field(self.nx, self.ny, self.nz)
            self.L.field_x = X[1:, 1:, 1:] - X[:-1, :-1, :-1]
            self.L.field_y = Y[1:, 1:, 1:] - Y[:-1, :-1, :-1]
            self.L.field_z = Z[1:, 1:, 1:] - Z[:-1, :-1, :-1]

            self.iA = Field(self.nx, self.ny, self.nz)
            self.iA.field_x = np.divide(1.0, self.L.field_y * self.L.field_z)
            self.iA.field_y = np.divide(1.0, self.L.field_x * self.L.field_z)
            self.iA.field_z = np.divide(1.0, self.L.field_x * self.L.field_y)

            # tilde grid ~G
            # self.itA = self.iA
            # self.tL = self.L

            self.tx = (self.x[1:] + self.x[:-1]) / 2
            self.ty = (self.y[1:] + self.y[:-1]) / 2
            self.tz = (self.z[1:] + self.z[:-1]) / 2

            self.tx = np.append(self.tx, self.tx[-1])
            self.ty = np.append(self.ty, self.ty[-1])
            self.tz = np.append(self.tz, self.tz[-1])

            tY, tX, tZ = np.meshgrid(self.ty, self.tx, self.tz)

            self.tL = Field(self.nx, self.ny, self.nz)
            self.tL.field_x = tX[1:, 1:, 1:] - tX[:-1, :-1, :-1]
            self.tL.field_y = tY[1:, 1:, 1:] - tY[:-1, :-1, :-1]
            self.tL.field_z = tZ[1:, 1:, 1:] - tZ[:-1, :-1, :-1]

            self.itA = Field(self.nx, self.ny, self.nz)
            aux = self.tL.field_y * self.tL.field_z
            self.itA.field_x = np.divide(
                1.0, aux, out=np.zeros_like(aux), where=aux != 0
            )
            aux = self.tL.field_x * self.tL.field_z
            self.itA.field_y = np.divide(
                1.0, aux, out=np.zeros_like(aux), where=aux != 0
            )
            aux = self.tL.field_x * self.tL.field_y
            self.itA.field_z = np.divide(
                1.0, aux, out=np.zeros_like(aux), where=aux != 0
            )
            del aux

            self.compute_areas(
                self.l_x,
                self.l_y,
                self.l_z,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.Sxy_red,
                self.Syz_red,
                self.Szx_red,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
            )
            self.mark_cells(
                self.l_x,
                self.l_y,
                self.l_z,
                self.nx,
                self.ny,
                self.nz,
                self.dx,
                self.dy,
                self.dz,
                self.Sxy,
                self.Syz,
                self.Szx,
                self.flag_int_cell_xy,
                self.flag_int_cell_yz,
                self.flag_int_cell_zx,
                self.Sxy_stab,
                self.Syz_stab,
                self.Szx_stab,
                self.flag_unst_cell_xy,
                self.flag_unst_cell_yz,
                self.flag_unst_cell_zx,
                self.flag_bound_cell_xy,
                self.flag_bound_cell_yz,
                self.flag_bound_cell_zx,
                self.flag_avail_cell_xy,
                self.flag_avail_cell_yz,
                self.flag_avail_cell_zx,
            )

    """
  Function to compute the length of the edges of the conformal grid.
    Inputs:
        - tol: an edge shorter than tol will be considered as of zero length
    """

    def compute_edges(self, tol=1e-8):
        # edges xy
        for kk in range(self.nz + 1):
            for ii in range(self.nx):
                for jj in range(self.ny + 1):
                    x_1 = ii * self.dx + self.xmin
                    y_1 = jj * self.dy + self.ymin
                    x_2 = (ii + 1) * self.dx + self.xmin
                    y_2 = jj * self.dy + self.ymin
                    z = kk * self.dz + self.zmin
                    # if point 1 is in conductor
                    if self.conductors.in_conductor(x_1, y_1, z):
                        # if point 2 is in conductor, length of the l_x[i, j] is zero
                        if self.conductors.in_conductor(x_2, y_2, z):
                            self.l_x[ii, jj, kk] = 0
                        # if point 2 is not in conductor, length of l_x[i, j]
                        # is the fractional length
                        else:
                            self.l_x[ii, jj, kk] = seg_length(
                                self.conductors.intersec_x(x_2, y_2, z),
                                y_2,
                                z,
                                x_2,
                                y_2,
                                z,
                            )
                    # if point 1 is not in conductor
                    else:
                        # if point 2 is in conductor, length of l_x[i, j] is the fractional length
                        if self.conductors.in_conductor(x_2, y_2, z):
                            self.l_x[ii, jj, kk] = seg_length(
                                x_1,
                                y_1,
                                z,
                                self.conductors.intersec_x(x_1, y_1, z),
                                y_1,
                                z,
                            )
                        # if point 2 is not in conductor, length of l_x[i, j] is dx
                        else:
                            self.l_x[ii, jj, kk] = self.dx

        for kk in range(self.nz + 1):
            for ii in range(self.nx + 1):
                for jj in range(self.ny):
                    x_1 = ii * self.dx + self.xmin
                    y_1 = jj * self.dy + self.ymin
                    x_3 = ii * self.dx + self.xmin
                    y_3 = (jj + 1) * self.dy + self.ymin
                    z = kk * self.dz + self.zmin
                    # if point 1 is in conductor
                    if self.conductors.in_conductor(x_1, y_1, z):
                        # if point 3 to the right is in conductor, length of the l_y[i, j] is zero
                        if self.conductors.in_conductor(x_3, y_3, z):
                            self.l_y[ii, jj, kk] = 0
                        # if point 3 is not in conductor, length of l_y[i, j]
                        # is the fractional length
                        else:
                            self.l_y[ii, jj, kk] = seg_length(
                                x_3,
                                self.conductors.intersec_y(x_3, y_3, z),
                                z,
                                x_3,
                                y_3,
                                z,
                            )
                    # if point 1 is not in conductor
                    else:
                        # if point 3 is in conductor, length of the l_y[i, j]
                        # is the fractional length
                        if self.conductors.in_conductor(x_3, y_3, z):
                            self.l_y[ii, jj, kk] = seg_length(
                                x_1,
                                y_1,
                                z,
                                x_1,
                                self.conductors.intersec_y(x_1, y_1, z),
                                z,
                            )
                        # if point 3 is not in conductor, length of l_y[i, j] is dy
                        else:
                            self.l_y[ii, jj, kk] = self.dy

        for jj in range(self.ny + 1):
            for ii in range(self.nx + 1):
                for kk in range(self.nz):
                    y = jj * self.dy + self.ymin
                    x_1 = ii * self.dx + self.xmin
                    z_1 = kk * self.dz + self.zmin
                    x_3 = ii * self.dx + self.xmin
                    z_3 = (kk + 1) * self.dz + self.zmin
                    # if point 1 is in conductor
                    if self.conductors.in_conductor(x_1, y, z_1):
                        # if point 3 to the right is in conductor, length of the l_y[i, j] is zero
                        if self.conductors.in_conductor(x_3, y, z_3):
                            self.l_z[ii, jj, kk] = 0
                        # if point 3 is not in conductor, length of l_y[i, j]
                        # is the fractional length
                        else:
                            self.l_z[ii, jj, kk] = seg_length(
                                x_3,
                                y,
                                self.conductors.intersec_z(x_3, y, z_3),
                                x_3,
                                y,
                                z_3,
                            )
                    # if point 1 is not in conductor
                    else:
                        # if point 3 is in conductor, length of the l_y[i, j]
                        # is the fractional length
                        if self.conductors.in_conductor(x_3, y, z_3):
                            self.l_z[ii, jj, kk] = seg_length(
                                x_1,
                                y,
                                z_1,
                                x_1,
                                y,
                                self.conductors.intersec_z(x_1, y, z_1),
                            )
                        # if point 3 is not in conductor, length of l_y[i, j] is dy
                        else:
                            self.l_z[ii, jj, kk] = self.dz

        # set to zero the length of very small cells
        if tol > 0.0:
            self.l_x /= self.dx
            self.l_y /= self.dy
            self.l_z /= self.dz

            low_values_flags = abs(self.l_x) < tol
            high_values_flags = abs(self.l_x - 1.0) < tol
            self.l_x[low_values_flags] = 0
            self.l_x[high_values_flags] = 1.0

            low_values_flags = abs(self.l_y) < tol
            high_values_flags = abs(self.l_y - 1.0) < tol
            self.l_y[low_values_flags] = 0
            self.l_y[high_values_flags] = 1.0

            low_values_flags = abs(self.l_z) < tol
            high_values_flags = abs(self.l_z - 1.0) < tol
            self.l_z[low_values_flags] = 0
            self.l_z[high_values_flags] = 1.0

            self.l_x *= self.dx
            self.l_y *= self.dy
            self.l_z *= self.dz

    """
  Function to compute the area of the cells of the conformal grid.
    """

    @staticmethod
    @jit(
        "(f8[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], i4, i4, i4, f8, f8, f8)",
        nopython=True,
    )
    def compute_areas(
        l_x, l_y, l_z, Sxy, Syz, Szx, Sxy_red, Syz_red, Szx_red, nx, ny, nz, dx, dy, dz
    ):
        for kk in range(nz + 1):
            compute_areas_2D(
                l_x[:, :, kk],
                l_y[:, :, kk],
                Sxy[:, :, kk],
                Sxy_red[:, :, kk],
                nx,
                ny,
                dx,
                dy,
            )

        for ii in range(nx + 1):
            compute_areas_2D(
                l_y[ii, :, :],
                l_z[ii, :, :],
                Syz[ii, :, :],
                Syz_red[ii, :, :],
                ny,
                nz,
                dy,
                dz,
            )

        for jj in range(ny + 1):
            compute_areas_2D(
                l_x[:, jj, :],
                l_z[:, jj, :],
                Szx[:, jj, :],
                Szx_red[:, jj, :],
                nx,
                nz,
                dx,
                dz,
            )

    """
  Function to mark wich cells are interior (int), require extension (unst), 
  are on the boundary(bound), are available for intrusion (avail)
    """

    @staticmethod
    @jit(
        "(f8[:,:,:], f8[:,:,:], f8[:,:,:], i4, i4, i4, f8, f8, f8, f8[:,:,:], f8[:,:,:], f8[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], f8[:,:,:], f8[:,:,:], f8[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:], b1[:,:,:])",
        nopython=True,
    )
    def mark_cells(
        l_x,
        l_y,
        l_z,
        nx,
        ny,
        nz,
        dx,
        dy,
        dz,
        Sxy,
        Syz,
        Szx,
        flag_int_cell_xy,
        flag_int_cell_yz,
        flag_int_cell_zx,
        Sxy_stab,
        Syz_stab,
        Szx_stab,
        flag_unst_cell_xy,
        flag_unst_cell_yz,
        flag_unst_cell_zx,
        flag_bound_cell_xy,
        flag_bound_cell_yz,
        flag_bound_cell_zx,
        flag_avail_cell_xy,
        flag_avail_cell_yz,
        flag_avail_cell_zx,
    ):
        for kk in range(nz + 1):
            mark_cells_2D(
                l_x[:, :, kk],
                l_y[:, :, kk],
                nx,
                ny,
                dx,
                dy,
                Sxy[:, :, kk],
                flag_int_cell_xy[:, :, kk],
                Sxy_stab[:, :, kk],
                flag_unst_cell_xy[:, :, kk],
                flag_bound_cell_xy[:, :, kk],
                flag_avail_cell_xy[:, :, kk],
            )

        for ii in range(nx + 1):
            mark_cells_2D(
                l_y[ii, :, :],
                l_z[ii, :, :],
                ny,
                nz,
                dy,
                dz,
                Syz[ii, :, :],
                flag_int_cell_yz[ii, :, :],
                Syz_stab[ii, :, :],
                flag_unst_cell_yz[ii, :, :],
                flag_bound_cell_yz[ii, :, :],
                flag_avail_cell_yz[ii, :, :],
            )

        for jj in range(ny + 1):
            mark_cells_2D(
                l_x[:, jj, :],
                l_z[:, jj, :],
                nx,
                nz,
                dx,
                dz,
                Szx[:, jj, :],
                flag_int_cell_zx[:, jj, :],
                Szx_stab[:, jj, :],
                flag_unst_cell_zx[:, jj, :],
                flag_bound_cell_zx[:, jj, :],
                flag_avail_cell_zx[:, jj, :],
            )

    """
  Function to compute the extension of the unstable cells
    """

    def compute_extensions(self):
        # breakpoint()
        for kk in range(self.nz + 1):
            # breakpoint()
            Grid2D.compute_extensions(
                nx=self.nx,
                ny=self.ny,
                S=self.Sxy[:, :, kk],
                flag_int_cell=self.flag_int_cell_xy[:, :, kk],
                S_stab=self.Sxy_stab[:, :, kk],
                S_enl=self.Sxy_enl[:, :, kk],
                S_red=self.Sxy_red[:, :, kk],
                flag_unst_cell=self.flag_unst_cell_xy[:, :, kk],
                flag_avail_cell=self.flag_avail_cell_xy[:, :, kk],
                flag_ext_cell=self.flag_ext_cell_xy[:, :, kk],
                flag_intr_cell=self.flag_intr_cell_xy[:, :, kk],
                borrowing=self.borrowing_xy[:, :, kk],
                kk=kk,
                l_verbose=False,
            )

        for ii in range(self.nx + 1):
            Grid2D.compute_extensions(
                nx=self.ny,
                ny=self.nz,
                S=self.Syz[ii, :, :],
                flag_int_cell=self.flag_int_cell_yz[ii, :, :],
                S_stab=self.Syz_stab[ii, :, :],
                S_enl=self.Syz_enl[ii, :, :],
                S_red=self.Syz_red[ii, :, :],
                flag_unst_cell=self.flag_unst_cell_yz[ii, :, :],
                flag_avail_cell=self.flag_avail_cell_yz[ii, :, :],
                flag_ext_cell=self.flag_ext_cell_yz[ii, :, :],
                borrowing=self.borrowing_yz[ii, :, :],
                flag_intr_cell=self.flag_intr_cell_yz[ii, :, :],
                kk=ii,
                l_verbose=False,
            )

        for jj in range(self.ny + 1):
            Grid2D.compute_extensions(
                nx=self.nx,
                ny=self.nz,
                S=self.Szx[:, jj, :],
                flag_int_cell=self.flag_int_cell_zx[:, jj, :],
                S_stab=self.Szx_stab[:, jj, :],
                S_enl=self.Szx_enl[:, jj, :],
                S_red=self.Szx_red[:, jj, :],
                flag_unst_cell=self.flag_unst_cell_zx[:, jj, :],
                flag_avail_cell=self.flag_avail_cell_zx[:, jj, :],
                flag_ext_cell=self.flag_ext_cell_zx[:, jj, :],
                flag_intr_cell=self.flag_intr_cell_zx[:, jj, :],
                borrowing=self.borrowing_zx[:, jj, :],
                kk=jj,
                l_verbose=False,
            )



import os
import sys

import numpy as np
import pyvista as pv

sys.path.append("../wakis")

import pytest
from scipy.constants import c
from tqdm import tqdm

from wakis import GridFIT3D, SolverFIT3D, WakeSolver
from wakis.sources import Beam


@pytest.mark.slow
class Test3Dplotting:
    img_folder = "tests/008_img/"

    def test_simulation(self, use_gpu, flag_offscreen):
        # ---------- Domain setup ---------

        # Geometry & Materials
        solid_1 = "tests/stl/007_vacuum_cavity.stl"
        solid_2 = "tests/stl/007_lossymetal_shell.stl"

        stl_solids = {"cavity": solid_1, "shell": solid_2}

        stl_materials = {
            "cavity": "vacuum",
            "shell": [30, 1.0, 30],  # [eps_r, mu_r, sigma[S/m]]
        }

        stl_colors = {"cavity": "tab:blue", "shell": "silver"}

        # Extract domain bounds from geometry
        solids = pv.read(solid_1) + pv.read(solid_2)
        xmin, xmax, ymin, ymax, ZMIN, ZMAX = solids.bounds

        # Number of mesh cells
        Nx = 60
        Ny = 60
        NZ = 140

        grid = GridFIT3D(
            xmin,
            xmax,
            ymin,
            ymax,
            ZMIN,
            ZMAX,
            Nx,
            Ny,
            NZ,
            use_mpi=False,  # Enables MPI subdivision of the domain
            stl_solids=stl_solids,
            stl_materials=stl_materials,
            stl_colors=stl_colors,
            stl_scale=1.0,
            stl_rotate=[0, 0, 0],
            stl_translate=[0, 0, 0],
            verbose=1,
        )

        # ------------ Beam source & Wake ----------------
        # Beam parameters
        sigmaz = 10e-2  # [m] -> 2 GHz
        q = 1e-9  # [C]
        beta = 1.0  # beam beta
        xs = 0.0  # x source position [m]
        ys = 1e-2  # y source position [m]
        ti = 3 * sigmaz / c  # injection time [s]

        beam = Beam(
            q=q, sigmaz=sigmaz, beta=beta, xsource=xs, ysource=ys, ti=ti
        )

        wake = WakeSolver(
            q=q, sigmaz=sigmaz, beta=beta, xsource=xs, ysource=ys, ti=ti
        )

        # ----------- Solver & Simulation ----------
        # boundary conditions
        bc_low = ["pec", "pec", "pec"]
        bc_high = ["pec", "pec", "pec"]

        # Solver setup
        global solver
        solver = SolverFIT3D(
            grid,
            wake,
            bc_low=bc_low,
            bc_high=bc_high,
            use_stl=True,
            use_mpi=False,  # Activate MPI
            bg="pec",  # Background material
            dtype=np.float32,
            use_gpu=use_gpu,
        )

        # -------------- Output folder ---------------------
        if not os.path.exists(self.img_folder):
            os.mkdir(self.img_folder)

        # -------------- Custom time loop  -----------------
        Nt = 1000
        for n in tqdm(range(Nt)):
            beam.update(solver, n * solver.dt)
            solver.one_step()

    def test_grid_inspect(self, flag_offscreen):
        # Plot grid and imported solids
        global solver
        pl = solver.grid.inspect(
            add_stl=["cavity", "shell"],
            stl_opacity=0.1,
            off_screen=flag_offscreen,
            anti_aliasing="ssaa",
        )
        if flag_offscreen:
            pl.screenshot(self.img_folder+'grid_inspect.png')
            # pl.export_html(self.img_folder + "grid_inspect.html") # requires trame   

    def test_grid_plot_solids(self, flag_offscreen):
        # Plot only imported solids
        global solver
        solver.grid.plot_solids(
            bounding_box=True,
            show_grid=False,
            opacity=1,
            specular=0.5,
            smooth_shading=False,
            off_screen=flag_offscreen,
        )

    def test_grid_stl_mask(self, flag_offscreen):
        # Plot STL solid masks in the grid
        global solver
        solver.grid.plot_stl_mask(
            stl_solid="cavity",
            cmap="viridis",
            add_stl="all",
            stl_opacity=0.5,
            smooth_shading=False,
            anti_aliasing="ssaa",
            ymax=0.0,
            off_screen=flag_offscreen,
        )

    def test_solver_inspect(self, flag_offscreen):
        # Plot imported solids and beam source and integraiton path
        global solver
        pl = solver.inspect(
            window_size=(1200, 800),
            off_screen=flag_offscreen,
            specular=0.0,
            opacity=1,
            inactive_opacity=0.1,
            add_silhouette=True,
        )
        if flag_offscreen:
            pl.screenshot(self.img_folder+'solver_inspect.png')
            # pl.export_html(self.img_folder + "solver_inspect.html") # requires trame

    def test_plot3D(self, flag_offscreen):
        # Plot Abs Electric field on domain
        global solver
        solver.plot3D(
            "E",
            component="z",
            cmap="rainbow",
            clim=[0, 500],
            add_stl=["cavity", "shell"],
            stl_opacity=0.1,
            clip_interactive=True,
            clip_normal="-y",
            title=self.img_folder + "Ez3d",
            off_screen=flag_offscreen,
        )

    def test_plot3DonSTL(self, flag_offscreen):
        # Plot Abs Electric field on STL solid `cavity`
        global solver
        solver.plot3DonSTL(
            "E",
            component="Abs",
            cmap="rainbow",
            clim=[0, 500],
            stl_with_field="cavity",
            field_opacity=1.0,
            stl_transparent="shell",
            stl_opacity=0.1,
            stl_colors="white",
            clip_plane=True,
            clip_normal="-y",
            clip_origin=[0, 0, 0],
            off_screen=flag_offscreen,
            zoom=1.2,
            title=self.img_folder + "EAbs3donSTL",
        )



import numpy as np

from wakis import GridFIT3D, SolverFIT3D, WakeSolver
from wakis.field import Field
from wakis.field_monitors import FieldMonitor


def _make_constant_field(Nx=2, Ny=2, Nz=2, values=(1.0, 2.0, 3.0)):
    """Create a Field with spatially constant components."""

    field = Field(Nx, Ny, Nz, dtype=float, use_ones=False, use_gpu=False)

    Ex = np.full((Nx, Ny, Nz), values[0], dtype=float)
    Ey = np.full((Nx, Ny, Nz), values[1], dtype=float)
    Ez = np.full((Nx, Ny, Nz), values[2], dtype=float)

    field.from_matrix(Ex, "x")
    field.from_matrix(Ey, "y")
    field.from_matrix(Ez, "z")

    return field


def test_initial_state():
    freqs = [1.0e9, 2.0e9]
    monitor = FieldMonitor(freqs)

    assert isinstance(monitor.frequencies, np.ndarray)
    assert np.allclose(monitor.frequencies, freqs)
    assert monitor.time_index == 0
    assert monitor.dt is None
    assert monitor.Ex_acc is None
    assert monitor.Ey_acc is None
    assert monitor.Ez_acc is None
    assert monitor.shape is None
    assert monitor.xp is None


def test_single_update_initializes_and_accumulates():
    dt = 1.0e-9
    field = _make_constant_field(Nx=2, Ny=2, Nz=2, values=(1.0, 2.0, 3.0))
    monitor = FieldMonitor([0.0])  # zero frequency -> unit phase

    monitor.update(field, dt)
    comps = monitor.get_components()

    assert monitor.dt == dt
    assert monitor.shape == (2, 2, 2)
    assert monitor.time_index == 1
    assert monitor.xp is field.xp

    for key in ("Ex", "Ey", "Ez"):
        arr = comps[key]
        assert arr is not None
        assert arr.shape == (1, 2, 2, 2)
        assert np.iscomplexobj(arr)

    Ex_expected = field.to_matrix("x")
    Ey_expected = field.to_matrix("y")
    Ez_expected = field.to_matrix("z")

    assert np.allclose(comps["Ex"][0], Ex_expected)
    assert np.allclose(comps["Ey"][0], Ey_expected)
    assert np.allclose(comps["Ez"][0], Ez_expected)


def test_multiple_updates_accumulate_linearly_for_zero_frequency():
    dt = 0.1
    n_steps = 5
    field = _make_constant_field(Nx=2, Ny=2, Nz=2, values=(1.0, 0.0, 0.0))
    monitor = FieldMonitor([0.0])  # exp(0) == 1 for all timesteps

    for _ in range(n_steps):
        monitor.update(field, dt)

    comps = monitor.get_components()

    assert monitor.time_index == n_steps
    assert monitor.dt == dt  # dt is fixed at first call

    Ex_expected = n_steps * field.to_matrix("x")
    assert np.allclose(comps["Ex"][0], Ex_expected)
    assert np.allclose(comps["Ey"], 0.0)
    assert np.allclose(comps["Ez"], 0.0)


def test_frequency_phase_accumulation_matches_geometric_series():
    # Choose f and dt such that the complex exponential alternates sign:
    # r = exp(-2j*pi*f*dt) with f*dt = 0.5 -> r = exp(-j*pi) = -1.
    f = 0.5
    dt = 1.0
    n_steps = 4

    field = _make_constant_field(Nx=1, Ny=1, Nz=1, values=(1.0, 0.0, 0.0))
    monitor = FieldMonitor([f])

    for _ in range(n_steps):
        monitor.update(field, dt)

    comps = monitor.get_components()
    Ex = comps["Ex"][0]

    # Geometric series: sum_{n=0}^{3} (-1)^n = 0
    assert np.allclose(Ex, 0.0, atol=1e-12)


def test_get_components_before_update_returns_none_accumulators():
    monitor = FieldMonitor([1.0])
    comps = monitor.get_components()

    assert comps["Ex"] is None
    assert comps["Ey"] is None
    assert comps["Ez"] is None


def test_field_monitor_integration_with_wakesolve():
    """Run a tiny wake simulation with a FieldMonitor attached.

    This exercises the integration between SolverFIT3D.wakesolve and
    FieldMonitor.update, ensuring that the monitor is updated during the
    time-stepping loop and produces non-empty accumulators.
    """

    # Small, uniform grid
    Nx = Ny = Nz = 4
    grid = GridFIT3D(
        xmin=0.0,
        xmax=1.0,
        ymin=0.0,
        ymax=1.0,
        zmin=0.0,
        zmax=1.0,
        Nx=Nx,
        Ny=Ny,
        Nz=Nz,
        verbose=0,
    )

    # Minimal wake configuration (no STL). Use a dedicated Ez file so that
    # the solver writes its Ez snapshots without interfering with other
    # tests or examples.
    ez_file = "tests/012_Ez_monitor.h5"
    wake = WakeSolver(
        q=1e-9,
        sigmaz=1e-2,
        beta=1.0,
        xsource=0.0,
        ysource=0.0,
        xtest=0.0,
        ytest=0.0,
        Ez_file=ez_file,
        save=False,
        results_folder="results/",
    )

    # For this integration test we only care that ``wakesolve`` runs its
    # time-stepping loop and calls the field monitor. The heavy wake
    # post-processing (HDF5 reads, impedance calculations, etc.) is
    # exercised elsewhere in the test suite, so we replace ``solve`` with
    # a lightweight no-op to keep this test fast and robust.
    wake.solve = lambda *args, **kwargs: None

    solver = SolverFIT3D(
        grid,
        wake,
        bc_low=["pec", "pec", "pec"],
        bc_high=["pec", "pec", "pec"],
        use_stl=False,
        bg="pec",
        dtype=np.float32,
        use_gpu=False,
    )

    # Attach a monitor at a single frequency
    monitor = FieldMonitor([1.0e9])

    # Run a very short wake simulation to keep the test fast
    wakelength = 0.1  # [m]
    solver.wakesolve(
        wakelength=wakelength,
        add_space=0,
        plot=False,
        use_field_monitor=True,
        field_monitor=monitor,
    )

    comps = monitor.get_components()

    # After the run, the monitor should have been initialized and updated
    assert monitor.dt is not None
    assert monitor.time_index > 0
    assert monitor.shape == (Nx, Ny, Nz)

    for key in ("Ex", "Ey", "Ez"):
        arr = comps[key]
        assert arr is not None
        assert arr.shape[0] == 1  # one frequency
        assert arr.shape[1:] == (Nx, Ny, Nz)

    # Clean up Ez file created for this test if it exists
    import os

    if os.path.exists(ez_file):
        os.remove(ez_file)



# copyright ################################# #
# This file is part of the wakis Package.     #
# Copyright (c) CERN, 2024.                   #
# ########################################### #

from pathlib import Path

from setuptools import find_packages, setup

# read version
version_file = Path(__file__).parent / "wakis/_version.py"
dd = {}
with open(version_file.absolute(), "r") as fp:
    exec(fp.read(), dd)
__version__ = dd["__version__"]

# read long_description
long_description = (Path(__file__).parent / "README.md").read_text(
    encoding="utf-8"
)

# read requirements.txt for extras_require
with open("requirements.txt") as f:
    notebook_required = f.read().splitlines()

setup(
    name="wakis",
    version=__version__,
    description="3D Electromagnetic Time-Domain wake and impedance solver",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://wakis.readthedocs.io/",
    author="Elena de la Fuente et al.",
    author_email="elena.de.la.fuente.garcia@cern.ch",
    license="Apache 2.0",
    download_url="https://pypi.python.org/pypi/wakis",
    project_urls={
        "Bug Tracker": "https://github.com/ImpedanCEI/wakis/issues",
        "Documentation": "https://wakis.readthedocs.io/en/latest/index.html",
        "Source Code": "https://github.com/ImpedanCEI/wakis/wakis",
    },
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    install_requires=[
        "numpy",
        "scipy",
        "pyvista",
        "h5py",
        "tqdm",
    ],
    extras_require={
        "gpu": ["cupy"],
        "all": ["pyvista[jupyter]", "jupyterlab", "ipympl", "bihc", "iddefix"],
        "notebook": notebook_required,
    },
    tests_require=["pytest", "pytest-cov"],
)
